class ClassPropertyDescriptor(object):
    """
    Wrapper which provides access to methods through property syntax
    """

    def __init__(self, f_get, f_set=None):
        self.f_get = f_get
        self.f_set = f_set

    def __get__(self, obj, klass=None):
        if klass is None:
            klass = type(obj)
        return self.f_get.__get__(obj, klass)()


def classproperty(func):
    """
    Decorator for properties of classes, similar to stdlib's `property` which is limited to properties of objects

    :param func: function to decorate
    :return: wrapper which provides access to methods through property syntax
    """
    if not isinstance(func, (classmethod, staticmethod)):
        func = classmethod(func)

    return ClassPropertyDescriptor(func)



import time
import sqlite3
import os
current_path = os.path.dirname(__file__)

def check_str(s, type):
    if not isinstance(s, str):
        raise Exception(type + ' not string!')
    elif s == '':
        raise Exception(type + ' is null!')

class ModelTrack(object):

    """
    :param nick_name:        str，用户名，多人使用下可起到数据隔离。
    :param project_name:     str，项目名称。
    :param project_remark:   str，项目备注，默认为空。
    项目名称如不存在会新建
    """
    def __init__(self, nick_name, project_name, project_remark=''):

        self.conn = sqlite3.connect(os.path.join(current_path, 'model_log.db'))

        self.nick_name = nick_name
        self.project_name = project_name
        self.project_remark = project_remark
        self.start_time = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
        self.is_add_model_data = True
        self.param_dict = {}
        self.model_name = ''
        self.remark = ''

        check_str(project_name, 'project_name')
        check_str(nick_name, 'nick_name')


    # 检查project name 是否存在
    def __is_exist_project_name(self, project_name):

        sql = "select 1 from project m where m.project_name = '%s' and m.nick_name='%s'"%(project_name, self.nick_name)
        project_table = self.conn.execute(sql).fetchall()

        if len(project_table) != 0:
            return True
        else:
            return False

    # 检查model name 是否存在
    def __is_exist_model_name(self, model_name, project_id):

        sql = "select 1 from sub_model  sm where sm.project_id = %d and sm.sub_model_name = '%s'" % (project_id, model_name)
        project_table = self.conn.execute(sql).fetchall()

        if len(project_table) != 0:
            return True
        else:
            return False


    """
    :param param_dict: dict，训练参数字典
    :param param_type: str，参数类型，例如：TF参数、word2vec参数等。
    """
    def add_param(self, param_dict, param_type):

        check_str(param_type, 'param_type')

        self.param_dict[param_type] = param_dict

    """
    :param model_name: str，模型名称
    """
    def add_model_name(self, model_name):

        check_str(model_name, 'model_name')

        self.model_name = model_name

    """
    :param remark: str，模型备注
    """
    def add_model_remark(self, remark):

        check_str(remark, 'remark')

        self.remark = remark


    """
    :param metric_name:  str，评估指标名称，可选择['train_loss', 'test_loss', 'test_acc', 'test_recall', 'test_precision', 'test_F1']
    :param metric_value: float，评估指标数值。
    :param epoch:        int，训练周期
    第一次调用该 API 时，会把以上设置的数据持久化到 SQLite 数据库。
    可以在每个 epoch 周期的最后使用该 API添加训练集和测试集的评估指标，web端会自动获取该数据。
    """
    def add_metric(self, metric_name, metric_value, epoch):

        check_str(metric_name, 'metric_name')

        if metric_name not in ['train_loss', 'test_loss', 'test_acc', 'test_recall', 'test_precision', 'test_F1']:
            raise Exception("Your metric_name：%s, not in ['train_loss', 'test_loss', 'test_acc', 'test_recall', 'test_precision', 'test_F1']" % (metric_name))

        if self.is_add_model_data:
            self.__add_model_data()
            self.is_add_model_data = False

        create_time = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
        sql = "insert into model_metric values (null, ?, ?, 'line', ?, ?, ?)"
        self.conn.execute(sql, (self.sub_model_id, metric_name, epoch, '%.4f'%(metric_value), create_time))

        self.conn.commit()


    """
    :param best_name:  str，最佳评估指标名称，
    :param best_value: float，最佳评估指标数值。
    :param best_epoch: int，训练周期
    添加当前模型训练中最佳的评估数据，一般放到模型训练的最后进行添加。
    """
    def add_best_result(self, best_name, best_value, best_epoch):

        check_str(best_name, 'best_name')

        create_time = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
        sql = "insert into best_result values (null, ?, ?, ?, ?, ?)"
        self.conn.execute(sql, (self.sub_model_id, best_name, '%.4f' % (best_value), best_epoch, create_time))

        self.conn.commit()


    # 检查model_name是否重复
    def __check_model_name(self, model_name, sub_model_count, project_id):

        if model_name == '':
            model_name = self.project_name + '_' + str(sub_model_count + 1)

        else:

            # 判断是否有model_name
            if self.__is_exist_model_name(model_name, project_id):
                model_name = model_name + '_' + str(sub_model_count + 1)
            else:
                model_name = self.model_name

        if self.__is_exist_model_name(model_name, project_id):
            return self.__check_model_name(model_name, sub_model_count, project_id)
        else:
            return model_name


    # 添加模型数据
    def __add_model_data(self):

        # 插入model
        if not self.__is_exist_project_name(self.project_name):
            sql = "insert into project values (null, ?, ?, ?, ?, 0)"
            self.conn.execute(sql, (self.project_name, self.project_remark, self.nick_name, self.start_time))
            self.conn.commit()

        sql = "select project_id from project m where m.project_name = '%s' and m.nick_name='%s'"%(self.project_name, self.nick_name)
        project_id = self.conn.execute(sql).fetchall()[0][0]

        sql = "select count(1) from sub_model sm where sm.project_id = %d"%(project_id)
        sub_model_count = self.conn.execute(sql).fetchall()[0][0]

        # 插入sub model
        model_name = self.__check_model_name(self.model_name, sub_model_count, project_id)
        sql = "insert into sub_model values (null, ?, ?, ?, ?, ?, ?, 0)"
        self.conn.execute(sql, (project_id, sub_model_count + 1, model_name, self.remark, self.nick_name, self.start_time))
        self.conn.commit()

        sql = "select sub_model_id from sub_model sm where sm.project_id = ? and sm.sub_model_name = ?"
        self.sub_model_id = self.conn.execute(sql, (project_id, model_name)).fetchall()[0][0]

        # 插入model param
        for param_type, value in self.param_dict.items():

            for param_name, param_value in value.items():
                sql = "insert into model_param values (null, ?, ?, ?, ?, ?)"
                self.conn.execute(sql, (self.sub_model_id, param_type, param_name, str(param_value), self.start_time))

    # db数据库初始化
    def init_db(self):

        sql_script = open(os.path.join(current_path, 'init_db.sql'), 'r', encoding='utf-8').read()

        self.conn.executescript(sql_script)

        self.conn.commit()

    """
    关闭 SQLite 数据库连接
    """
    def close(self):
        self.conn.close()


if __name__ == '__main__':



    model_log = ModelLog('test项目', 'test备注')
    '''
    # 训练参数
    tf_param = {}
    tf_param['model'] = 'BILSTM_Attention'
    tf_param['optimizer'] = 'adam'
    tf_param['num_classes'] = 10
    tf_param['lr'] = 0.001
    tf_param['training_steps'] = 1000000
    tf_param['display_step'] = 1
    tf_param['batch_size'] = 100
    tf_param['num_hidden'] = 128
    tf_param['embedding_dim'] = 128
    tf_param['drop_out'] = 0.5
    model_log.add_model_name('bilstm_crf')
    model_log.add_model_remark('bilstm_crf备注')
    model_log.add_param(tf_param, 'tf_param')
    for i in range(1, 21):
        model_log.add_diagram('train_loss', 25.0 - i, i)
        model_log.add_diagram('test_loss', 24.0 - i, i)
        model_log.add_diagram('test_acc', 0.3 + 0.02 * i, i)
    model_log.add_best_result('best_loss', 2.34, 20)
    model_log.add_best_result('best_acc', 0.99, 20)
    
    '''

    #model_log.init_db()



from abc import abstractmethod
from typing import Any, Dict, Iterable, List, Optional, Type, TypeVar

from pyjackson import dumps, loads
from sqlalchemy import Column, DateTime, ForeignKey, Integer, String, Text, UniqueConstraint, Float
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship

from algolink.core.objects import DatasetType
from algolink.core.objects.artifacts import ArtifactCollection
from algolink.core.objects.core import (Buildable, EvaluationResults, EvaluationSet, Image, Model, Pipeline,
                                        PipelineStep, Project, RuntimeEnvironment, RuntimeInstance, Task, 
                                        Experiment, ModelMetric, ModelParam, BestResult)
from algolink.core.objects.dataset_source import DatasetSource
from algolink.core.objects.metric import Metric
from algolink.core.objects.requirements import Requirements

SQL_OBJECT_FIELD = '_sqlalchemy_object'


def json_column():
    return Column(Text)


def safe_loads(payload, as_class):
    return loads(payload, Optional[as_class])


def sqlobject(obj):
    return getattr(obj, SQL_OBJECT_FIELD, None)


def update_attrs(obj, **attrs):
    for name, value in attrs.items():
        setattr(obj, name, value)


T = TypeVar('T')
S = TypeVar('S', bound='Attaching')


class Attaching:
    id = ...
    name = ...

    def attach(self, obj):
        setattr(obj, SQL_OBJECT_FIELD, self)
        return obj

    @classmethod
    def from_obj(cls: Type[S], obj: T, new=False) -> S:
        kwargs = cls.get_kwargs(obj)
        existing = sqlobject(obj)
        if not new and existing is not None:
            update_attrs(existing, **kwargs)
            return existing
        return cls(**kwargs)

    @classmethod
    @abstractmethod
    def get_kwargs(cls, obj: T) -> dict:
        pass  # pragma: no cover

    @abstractmethod
    def to_obj(self) -> T:
        pass  # pragma: no cover


Base = declarative_base()


class SProject(Base, Attaching):
    __tablename__ = 'projects'
    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String, unique=True, nullable=False)
    author = Column(String, unique=False, nullable=False)
    creation_date = Column(DateTime, unique=False, nullable=False)

    tasks: Iterable['STask'] = relationship("STask", back_populates="project")

    def to_obj(self) -> Project:
        p = Project(self.name, id=self.id, author=self.author, creation_date=self.creation_date)
        for task in self.tasks:
            p._tasks.add(task.to_obj())
        return self.attach(p)

    @classmethod
    def get_kwargs(cls, project: Project) -> dict:
        return dict(id=project.id,
                    name=project.name,
                    author=project.author,
                    creation_date=project.creation_date,
                    tasks=[STask.from_obj(t) for t in project.tasks.values()])


class STask(Base, Attaching):
    __tablename__ = 'tasks'

    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String, unique=False, nullable=False)
    author = Column(String, unique=False, nullable=False)
    creation_date = Column(DateTime, unique=False, nullable=False)
    project_id = Column(Integer, ForeignKey('projects.id'), nullable=False)

    project = relationship("SProject", back_populates="tasks")
    models: Iterable['SModel'] = relationship("SModel", back_populates="task")
    pipelines: Iterable['SPipeline'] = relationship("SPipeline", back_populates='task')
    images: Iterable['SImage'] = relationship("SImage", back_populates='task')
    
    experiments: Iterable['TExperiment'] = relationship("TExperiment", back_populates='task')

    datasets = Column(Text)
    metrics = Column(Text)
    evaluation_sets = Column(Text)

    __table_args__ = (UniqueConstraint('name', 'project_id', name='tasks_name_and_ref'),)

    def to_obj(self) -> Task:
        task = Task(id=self.id,
                    name=self.name,
                    author=self.author,
                    creation_date=self.creation_date,
                    project_id=self.project_id,
                    datasets=safe_loads(self.datasets, Dict[str, DatasetSource]),
                    metrics=safe_loads(self.metrics, Dict[str, Metric]),
                    evaluation_sets=safe_loads(self.evaluation_sets, Dict[str, EvaluationSet]))
        for model in self.models:
            task._models.add(model.to_obj())

        for pipeline in self.pipelines:
            task._pipelines.add(pipeline.to_obj())

        for image in self.images:
            task._images.add(image.to_obj())

        for experiment in self.experiments:
            task._experiments.add(experiment.to_obj())
        return self.attach(task)

    @classmethod
    def get_kwargs(cls, task: Task) -> dict:
        return dict(id=task.id,
                    name=task.name,
                    author=task.author,
                    creation_date=task.creation_date,
                    project_id=task.project_id,
                    models=[SModel.from_obj(m) for m in task.models.values()],
                    images=[SImage.from_obj(i) for i in task.images.values()],
                    pipelines=[SPipeline.from_obj(p) for p in task.pipelines.values()],
                    datasets=dumps(task.datasets),
                    metrics=dumps(task.metrics),
                    evaluation_sets=dumps(task.evaluation_sets))


class SModel(Base, Attaching):
    __tablename__ = 'models'

    id = Column(Integer, primary_key=True, autoincrement=True)

    name = Column(String, unique=False, nullable=False)
    author = Column(String, unique=False, nullable=False)
    creation_date = Column(DateTime, unique=False, nullable=False)
    wrapper = Column(Text)

    artifact = Column(Text)
    requirements = Column(Text)
    description = Column(Text)
    params = Column(Text)
    task_id = Column(Integer, ForeignKey('tasks.id'), nullable=False)
    task = relationship("STask", back_populates="models")

    evaluations = Column(Text)
    __table_args__ = (UniqueConstraint('name', 'task_id', name='models_name_and_ref'),)

    def to_obj(self) -> Model:
        model = Model(name=self.name,
                      wrapper_meta=safe_loads(self.wrapper, dict),
                      author=self.author,
                      creation_date=self.creation_date,
                      artifact=safe_loads(self.artifact, ArtifactCollection),
                      requirements=safe_loads(self.requirements, Requirements),
                      description=self.description,
                      params=safe_loads(self.params, Dict[str, Any]),
                      id=self.id,
                      task_id=self.task_id,
                      evaluations=safe_loads(self.evaluations, Dict[str, EvaluationResults]))
        return self.attach(model)

    @classmethod
    def get_kwargs(cls, model: Model) -> dict:
        return dict(id=model.id,
                    name=model.name,
                    author=model.author,
                    creation_date=model.creation_date,
                    wrapper=dumps(model.wrapper_meta),
                    artifact=dumps(model.artifact),
                    requirements=dumps(model.requirements),
                    description=model.description,
                    params=dumps(model.params),
                    task_id=model.task_id,
                    evaluations=dumps(model.evaluations))


class SPipeline(Base, Attaching):
    __tablename__ = 'pipelines'

    id = Column(Integer, primary_key=True, autoincrement=True)

    name = Column(String, unique=False, nullable=False)
    author = Column(String, unique=False, nullable=False)
    creation_date = Column(DateTime, unique=False, nullable=False)
    steps = Column(Text)

    input_data = Column(Text)
    output_data = Column(Text)

    task_id = Column(Integer, ForeignKey('tasks.id'), nullable=False)
    task = relationship("STask", back_populates="pipelines")

    evaluations = Column(Text)
    __table_args__ = (UniqueConstraint('name', 'task_id', name='pipelines_name_and_ref'),)

    def to_obj(self) -> Pipeline:
        pipeline = Pipeline(name=self.name,
                            steps=safe_loads(self.steps, List[PipelineStep]),
                            input_data=safe_loads(self.input_data, DatasetType),
                            output_data=safe_loads(self.output_data, DatasetType),
                            author=self.author,
                            creation_date=self.creation_date,
                            id=self.id,
                            task_id=self.task_id,
                            evaluations=safe_loads(self.evaluations, EvaluationResults))
        return self.attach(pipeline)

    @classmethod
    def get_kwargs(cls, pipeline: Pipeline) -> dict:
        return dict(id=pipeline.id,
                    name=pipeline.name,
                    author=pipeline.author,
                    creation_date=pipeline.creation_date,
                    steps=dumps(pipeline.steps),
                    input_data=dumps(pipeline.input_data),
                    output_data=dumps(pipeline.output_data),
                    task_id=pipeline.task_id,
                    evaluations=dumps(pipeline.evaluations))


class SImage(Base, Attaching):
    __tablename__ = 'images'

    id = Column(Integer, primary_key=True, autoincrement=True)

    name = Column(String, unique=False, nullable=False)
    author = Column(String, unique=False, nullable=False)
    creation_date = Column(DateTime, unique=False, nullable=False)

    task_id = Column(Integer, ForeignKey('tasks.id'), nullable=False)
    task = relationship("STask", back_populates="images")

    environment_id = Column(Integer, ForeignKey('environments.id'), nullable=False)

    params = Column(Text)
    source = Column(Text)

    __table_args__ = (UniqueConstraint('name', 'task_id', name='image_name_and_ref'),)

    def to_obj(self) -> Image:
        image = Image(name=self.name,
                      author=self.author,
                      creation_date=self.creation_date,
                      id=self.id,
                      task_id=self.task_id,
                      params=safe_loads(self.params, Image.Params),
                      source=safe_loads(self.source, Buildable),
                      environment_id=self.environment_id)
        return self.attach(image)

    @classmethod
    def get_kwargs(cls, image: Image) -> dict:
        return dict(id=image.id,
                    name=image.name,
                    author=image.author,
                    creation_date=image.creation_date,
                    task_id=image.task_id,
                    params=dumps(image.params),
                    source=dumps(image.source),
                    environment_id=image.environment_id)


class SRuntimeEnvironment(Base, Attaching):
    __tablename__ = 'environments'

    id = Column(Integer, primary_key=True, autoincrement=True)

    name = Column(String, unique=True, nullable=False)
    author = Column(String, unique=False, nullable=False)
    creation_date = Column(DateTime, unique=False, nullable=False)

    params = Column(Text)

    def to_obj(self) -> RuntimeEnvironment:
        environment = RuntimeEnvironment(
            name=self.name,
            author=self.author,
            creation_date=self.creation_date,
            id=self.id,
            params=safe_loads(self.params, RuntimeEnvironment.Params))
        return self.attach(environment)

    @classmethod
    def get_kwargs(cls, environment: RuntimeEnvironment) -> dict:
        return dict(id=environment.id,
                    name=environment.name,
                    author=environment.author,
                    creation_date=environment.creation_date,
                    params=dumps(environment.params))


class SRuntimeInstance(Base, Attaching):
    __tablename__ = 'instances'

    id = Column(Integer, primary_key=True, autoincrement=True)

    name = Column(String, unique=False, nullable=False)
    author = Column(String, unique=False, nullable=False)
    creation_date = Column(DateTime, unique=False, nullable=False)

    image_id = Column(Integer, ForeignKey('images.id'), nullable=False)
    environment_id = Column(Integer, ForeignKey('environments.id'), nullable=False)

    params = Column(Text)

    __table_args__ = (UniqueConstraint('name', 'image_id', 'environment_id', name='instance_name_and_ref'),)

    def to_obj(self) -> RuntimeInstance:
        instance = RuntimeInstance(
            name=self.name,
            author=self.author,
            creation_date=self.creation_date,
            id=self.id,
            image_id=self.image_id,
            environment_id=self.environment_id,
            params=safe_loads(self.params, RuntimeInstance.Params))
        return self.attach(instance)

    @classmethod
    def get_kwargs(cls, instance: RuntimeInstance) -> dict:
        return dict(id=instance.id,
                    name=instance.name,
                    author=instance.author,
                    creation_date=instance.creation_date,
                    image_id=instance.image_id,
                    environment_id=instance.environment_id,
                    params=dumps(instance.params))


class TExperiment(Base, Attaching):
    __tablename__ = 'experiments'

    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String, unique=False, nullable=False)
    author = Column(String, unique=False, nullable=False)
    experiment_remark = Column(String, unique=False, nullable=False)
    experiment_sequence = Column(Integer, nullable=False)
    del_flag = Column(Integer, default = 0)

    creation_date = Column(DateTime, unique=False, nullable=False)
    task_id = Column(Integer, ForeignKey('tasks.id'), nullable=False)

    task = relationship("STask", back_populates="experiments")
    modelmetrics: Iterable['TModelmetric'] = relationship("TModelmetric", back_populates="experiment")
    modelparams: Iterable['TModelparam'] = relationship("TModelparam", back_populates="experiment")
    bestresult: Iterable['TBestresult'] = relationship("TBestresult", back_populates="experiment")


    __table_args__ = (UniqueConstraint('name', 'task_id', name='experiment_name_and_ref'),)

    def to_obj(self) -> Experiment:
        experiment = Experiment(
                        id=self.id,
                        name=self.name,
                        author=self.author,
                        creation_date=self.creation_date,
                        task_id=self.task_id,
                        experiment_remark=self.experiment_remark,
                        experiment_sequence=self.experiment_sequence)
        #for model_metric in self.modelmetrics:
        #    experiment._modelmetrics.add(model_metric.to_obj())

        #for model_param in self.modelparams:
        #    experiment._modelparams.add(model_param.to_obj())

        #for best_result in self.bestresult:
        #    experiment._bestresult.add(best_result.to_obj())
        return self.attach(experiment)

    @classmethod
    def get_kwargs(cls, experiment: Experiment) -> dict:
        return dict(id=experiment.id,
                    name=experiment.name,
                    author=experiment.author,
                    creation_date=experiment.creation_date,
                    task_id=experiment.task_id,
                    experiment_remark=experiment.experiment_remark,
                    experiment_sequence=experiment.experiment_sequence,
                    del_flag=experiment.del_flag)

class TModelmetric(Base, Attaching):
    __tablename__ = 'model_metrics'

    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String, unique=False, nullable=False)
    author = Column(String, unique=False, nullable=False)
    metric_type = Column(String, unique=False, nullable=False)
    metric_value = Column(Float, nullable=False)
    epoch = Column(Integer, nullable=False)

    creation_date = Column(DateTime, unique=False, nullable=False)
    experiment_id = Column(Integer, ForeignKey('experiments.id'), nullable=False)

    experiment = relationship("TExperiment", back_populates="modelmetrics")

    def to_obj(self) -> ModelMetric:
        modelmetric = ModelMetric(id=self.id,
                        name=self.name,
                        author=self.author,
                        creation_date=self.creation_date,
                        metric_type=self.metric_type,
                        metric_value=self.metric_value,
                        epoch=self.epoch,
                        experiment_id=self.experiment_id)
        return self.attach(modelmetric)

    @classmethod
    def get_kwargs(cls, model_metric: ModelMetric) -> dict:
        return dict(id=model_metric.id,
                    name=model_metric.name,
                    author=model_metric.author,
                    creation_date=model_metric.creation_date,
                    experiment_id=model_metric.experiment_id,
                    metric_type = model_metric.metric_type,
                    metric_value = model_metric.metric_value,
                    epoch = model_metric.epoch)
    
class TModelparam(Base, Attaching):
    __tablename__ = 'model_params'

    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String, unique=False, nullable=False)
    author = Column(String, unique=False, nullable=False)
    param_type = Column(String, unique=False, nullable=False)
    param_value = Column(String, nullable=False)

    creation_date = Column(DateTime, unique=False, nullable=False)
    experiment_id = Column(Integer, ForeignKey('experiments.id'), nullable=False)

    experiment = relationship("TExperiment", back_populates="modelparams")

    def to_obj(self) -> ModelParam:
        model_param = ModelParam(id=self.id,
                    name=self.name,
                    author=self.author,
                    creation_date=self.creation_date,
                    experiment_id=self.experiment_id,
                    param_type=self.param_type,
                    param_value=self.param_value)
        return self.attach(model_param)

    @classmethod
    def get_kwargs(cls, model_param: ModelParam) -> dict:
        return dict(id=model_param.id,
                    name=model_param.name,
                    author=model_param.author,
                    creation_date=model_param.creation_date,
                    experiment_id=model_param.experiment_id,
                    param_type = model_param.param_type,
                    param_value = model_param.param_value)
    
class TBestresult(Base, Attaching):
    __tablename__ = 'best_result'

    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String, unique=False, nullable=False)
    author = Column(String, unique=False, nullable=False)
    best_value = Column(Float, nullable=False)
    best_epoch = Column(Integer, nullable=False)

    creation_date = Column(DateTime, unique=False, nullable=False)
    experiment_id = Column(Integer, ForeignKey('experiments.id'), nullable=False)

    experiment = relationship("TExperiment", back_populates="bestresult")

    def to_obj(self) -> BestResult:
        best_result = BestResult(id=self.id,
                    name=self.name,
                    author=self.author,
                    creation_date=self.creation_date,
                    best_value=self.best_value,
                    best_epoch=self.best_epoch,
                    experiment_id=self.experiment_id)

        return self.attach(best_result)

    @classmethod
    def get_kwargs(cls, best_result: BestResult) -> dict:
        return dict(id=best_result.id,
                    name=best_result.name,
                    author=best_result.author,
                    creation_date=best_result.creation_date,
                    best_value=best_result.best_value,
                    best_epoch = best_result.best_epoch,
                    experiment_id = best_result.experiment_id)


import logging.config

from ..config import Logging

LOG_LEVEL = logging._nameToLevel.get(Logging.LOG_LEVEL, logging.DEBUG)

logging_config = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'standard': {
            'format': '%(asctime)s [%(levelname)s] %(name)s: %(message)s'
        },
    },
    'handlers': {
        'default': {
            'level': LOG_LEVEL,
            'formatter': 'standard',
            'class': 'logging.StreamHandler',
            'stream': 'ext://sys.stdout',
        },
    },
    'loggers': {
        'algolink': {
            'handlers': ['default'],
            'level': LOG_LEVEL,
        },
        'algolink_runtime': {
            'handlers': ['default'],
            'level': LOG_LEVEL
        }
    }
}

logging.config.dictConfig(logging_config)

logger = logging.getLogger('algolink')
rlogger = logging.getLogger('algolink_runtime')



from .base import LOADER_ENV, ProviderBase, PythonProvider, SERVER_ENV
from .ml_model import MLModelProvider
from .ml_model_multi import MLModelMultiProvider
from .pipeline import PipelineProvider
__all__ = ['LOADER_ENV', 'ProviderBase', 'PythonProvider', 'SERVER_ENV', 'MLModelProvider', 'MLModelMultiProvider',
           'PipelineProvider']



from .artifacts import ArtifactCollection
from .core import (Image, Model, Pipeline, PipelineStep, Project, RuntimeEnvironment, RuntimeInstance, Task, 
                   Experiment, ModelMetric, ModelParam, BestResult)
from .dataset_type import DatasetType
from .requirements import Requirement, Requirements
from .wrapper import ModelIO, ModelWrapper

__all__ = ['Project', 'Requirements', 'Requirement', 'ArtifactCollection', 'ModelWrapper', 'Task', 'Image', 'Model',
           'DatasetType', 'RuntimeEnvironment', 'RuntimeInstance', 'ModelIO', 'Pipeline', 'PipelineStep',
           'Experiment','ModelMetric','ModelParam,' 'BestResult']



import glob
import os
import shutil
import typing

from algolink.core.errors import ArtifactExistsError, NoSuchArtifactError
from algolink.core.objects.artifacts import ArtifactCollection, Blob, Blobs, LocalFileBlob
from algolink.repository.artifact import ArtifactRepository
from algolink.utils.fs import get_lib_path
from algolink.utils.log import logger


class LocalArtifactRepository(ArtifactRepository):
    """
    :class:`.ArtifactRepository` implementation which stores artifacts in a local file system as directory

    :param: path: path to directory where artifacts are to be stored,
      if `None` "local_storage" directory in AlgoLink distribution is used
    """
    type = 'local'

    def __init__(self, path: str = None):
        self.path = os.path.abspath(path or get_lib_path('local_storage'))

    def get_artifact(self, artifact_type, artifact_id: str) -> ArtifactCollection:
        artifact_id = f'{artifact_type}/{artifact_id}'
        path = os.path.join(self.path, artifact_id)
        if not os.path.exists(path):
            raise NoSuchArtifactError(artifact_id, self)
        return Blobs({
            os.path.relpath(file, path): LocalFileBlob(os.path.join(self.path, file)) for file in
            glob.glob(os.path.join(path, '**'), recursive=True) if os.path.isfile(file)
        })

    def push_artifact(self, artifact_type, artifact_id: str, blobs: typing.Dict[str, Blob]) -> ArtifactCollection:
        artifact_id = f'{artifact_type}/{artifact_id}'
        path = os.path.join(self.path, artifact_id)
        if os.path.exists(path):
            raise ArtifactExistsError(artifact_id, self)

        os.makedirs(path, exist_ok=True)
        result = {}
        for filepath, blob in blobs.items():
            join = os.path.join(path, filepath)
            os.makedirs(os.path.dirname(join), exist_ok=True)
            logger.debug('Writing artifact %s to %s', blob, join)
            blob.materialize(join)
            result[filepath] = LocalFileBlob(join)
        return Blobs(result)

    def delete_artifact(self, artifact_type, artifact_id: str):
        artifact_id = f'{artifact_type}/{artifact_id}'
        path = os.path.join(self.path, artifact_id)
        if not os.path.exists(path):
            raise NoSuchArtifactError(artifact_id, self)
        logger.debug('Removing artifact %s', path)
        shutil.rmtree(path)



import contextlib
import os
import tempfile
import typing

import xgboost

from algolink.core.analyzer.base import TypeHookMixin
from algolink.core.analyzer.model import BindingModelHook
from algolink.core.objects.artifacts import Blobs, LocalFileBlob
from algolink.core.objects.wrapper import FilesContextManager, LibModelWrapperMixin, ModelIO, ModelWrapper


class XGBoostModelIO(ModelIO):
    """
    :class:`~.ModelIO` implementation for XGBoost models
    """
    model_path = 'model.xgb'

    @contextlib.contextmanager
    def dump(self, model: xgboost.Booster) -> FilesContextManager:
        with tempfile.TemporaryDirectory(prefix='algolink_xgboost_dump') as f:
            path = os.path.join(f, self.model_path)
            model.save_model(path)
            yield Blobs({self.model_path: LocalFileBlob(path)})

    def load(self, path):
        model = xgboost.Booster()
        model.load_model(os.path.join(path, self.model_path))
        return model


class XGBoostModelWrapper(LibModelWrapperMixin):
    """
    :class:`~.ModelWrapper` implementation for XGBoost models
    """
    libraries = [xgboost]

    def __init__(self):
        super().__init__(XGBoostModelIO())

    def _exposed_methods_mapping(self) -> typing.Dict[str, str]:
        return {
            'predict': '_predict'
        }

    @ModelWrapper.with_model
    def _predict(self, data):
        if not isinstance(data, xgboost.DMatrix):
            data = xgboost.DMatrix(data)
        return self.model.predict(data)


class XGBoostModelHook(BindingModelHook, TypeHookMixin):
    """
    :class:`.ModelHook` implementation for `xgboost.Booster` objects
    """
    valid_types = [xgboost.Booster]

    def _wrapper_factory(self) -> ModelWrapper:
        return XGBoostModelWrapper()



import os
from typing import Dict

from pyjackson import read

from algolink.core.objects import Model, Pipeline
from algolink.runtime.interface import Interface, expose
from algolink.runtime.interface.base import InterfaceLoader
from algolink.utils.log import rlogger

MODEL_BIN_PATH = 'model_dump'
PIPELINE_META_PATH = 'pipeline.json'


class PipelineMeta:
    def __init__(self, pipeline: Pipeline,
                 models: Dict[str, Model]):
        self.pipeline = pipeline
        self.models = models


def pipeline_interface(pipeline_meta: Pipeline):
    """
    Creates an interface from given pipeline with `run` method
    Method signature is determined via metadata associated with given pipeline.

    :param pipeline_meta: pipeline to create interface for
    :return: instance of :class:`.Interface` implementation
    """

    rlogger.debug('Creating interface for pipeline %s', pipeline_meta)

    class PipelineInterface(Interface):
        def __init__(self, pipeline):
            self.pipeline = pipeline

        @expose
        def run(self, data: pipeline_meta.input_data) -> pipeline_meta.output_data:
            rlogger.debug('running pipeline given %s', data)
            output_data = self.pipeline.run(data)
            rlogger.debug('run returned: %s', output_data)
            return pipeline_meta.output_data.serialize(output_data)

    return PipelineInterface(pipeline_meta)


class PipelineLoader(InterfaceLoader):
    """
    Implementation of :class:`.InterfaceLoader` which loads a pipeline via PyJackson and wraps it into an interface
    """

    def load(self) -> Interface:
        meta = read(PIPELINE_META_PATH, PipelineMeta)
        for name, model in meta.models.items():
            model.wrapper.load(os.path.join(MODEL_BIN_PATH, name))
        meta.pipeline.models = meta.models
        return pipeline_interface(meta.pipeline)



from .dataset import TorchTensorDatasetType, TorchTensorHook
from .model import TorchModelHook, TorchModelWrapper

__all__ = ['TorchTensorDatasetType', 'TorchTensorHook', 'TorchModelHook', 'TorchModelWrapper']



import argparse
from abc import abstractmethod
from typing import Any, Callable, Dict, Type

from everett import NO_VALUE, ConfigurationMissingError
from everett.manager import ConfigManager, ConfigOSEnv, ListOf, generate_uppercase_key

_config = ConfigManager([])


class ConfigEnv:
    register = True
    on_top = True

    @abstractmethod
    def get(self, key, namespace=None):
        raise NotImplementedError  # pragma: no cover

    def __init_subclass__(cls: Type['ConfigEnv']):
        if cls.register:
            inst = cls()
            if cls.on_top:
                _config.envs.insert(0, inst)
            else:
                _config.envs.append(inst)


class _ConfigArgParseEnv(ConfigEnv):
    on_top = False

    def __init__(self):
        self.cache = dict()

    def get(self, key, namespace=None):
        name = generate_uppercase_key(key, namespace).lower()
        if name in self.cache:
            return self.cache[name]
        parser = argparse.ArgumentParser()
        parser.add_argument(f'--{name}')
        args, _ = parser.parse_known_args()
        res = getattr(args, name) or NO_VALUE
        self.cache[name] = res
        return res


class _NamespacedOSEnv(ConfigOSEnv, ConfigEnv):
    namespace = 'ALGOLINK'

    def get(self, key, namespace=None):
        return super(_NamespacedOSEnv, self).get(key, namespace or self.namespace)


class Param:
    def __init__(self, key, namespace=None, default=NO_VALUE,
                 alternate_keys=NO_VALUE, doc='', parser: Callable = str, raise_error=True,
                 raw_value=False):
        self.key = key
        self.namespace = namespace
        self.default = default
        self.alternate_keys = alternate_keys
        self.doc = doc
        self.parser = parser
        self.raise_error = raise_error
        self.raw_value = raw_value

    def __get__(self, instance: 'Config', owner: Type['Config']):
        if instance is None:
            return self
        return _config(key=self.key, namespace=self.namespace or instance.namespace,
                       default=self.default, alternate_keys=self.alternate_keys,
                       doc=self.doc, parser=self.parser,
                       raise_error=self.raise_error, raw_value=self.raw_value)


class _ConfigMeta(type):
    def __new__(mcs, name, bases, namespace):
        meta = super().__new__(mcs, name + 'Meta', (mcs,) + bases, namespace)
        res = super().__new__(meta, name, bases, {})
        return res


class Config(metaclass=_ConfigMeta):
    namespace = None

    @classmethod
    def _try__get__(cls, value, default):
        try:
            return value.__get__(cls, type(cls))
        except ConfigurationMissingError:
            return default

    @classmethod
    def get_params(cls) -> Dict[str, Any]:
        return {
            name: cls._try__get__(value, '--NOT-SET--')
            for name, value in cls.__dict__.items() if isinstance(value, Param)
        }

    @classmethod
    def log_params(cls):
        from .utils.log import logger
        logger.debug('%s environment:', cls.__name__)
        for name, value in cls.get_params().items():
            logger.debug('%s: %s', name, value)


class Core(Config):
    DEBUG = Param('debug', default='false', doc='turn debug on', parser=bool)
    ADDITIONAL_EXTENSIONS = Param('extensions', default='',
                                  doc='comma-separated list of additional algolink extensions to load',
                                  parser=ListOf(str),
                                  raise_error=False)
    AUTO_IMPORT_EXTENSIONS = Param('auto_import_extensions', default='true',
                                   doc='Set to true to automatically load available extensions on algolink import',
                                   parser=bool)
    RUNTIME = Param('runtime', default='false', doc='is this instance a runtime', parser=bool)


class Logging(Config):
    LOG_LEVEL = Param('log_level', default='INFO' if not Core.DEBUG else 'DEBUG',
                      doc='Logging level for algolink',
                      parser=str)


class Runtime(Config):
    SERVER = Param('server', doc='server for runtime')
    LOADER = Param('loader', doc='interface loader for runtime')


if Core.DEBUG:
    Logging.log_params()
    Core.log_params()
    Runtime.log_params()



from typing import Mapping

import tensorflow as tf
from pyjackson.errors import SerializationError

from algolink.core.analyzer.base import CanIsAMustHookMixin
from algolink.core.analyzer.dataset import DatasetAnalyzer, DatasetHook
from algolink.core.objects.dataset_type import DatasetType, DictDatasetType
from algolink.core.objects.requirements import InstallableRequirement, Requirements


class FeedDictDatasetType(DictDatasetType):
    """
    :class:`~algolink.core.objects.DatasetType` implementation for tensorflow feed dict argument
    """

    @classmethod
    def from_feed_dict(cls, feed_dict):
        """
        Factory method to create :class:`FeedDictDatasetType` from feed dict

        :param feed_dict: feed dict
        :return: :class:`FeedDictDatasetType` instance
        """
        types = {}
        for k, v in feed_dict.items():
            types[cls.get_key(k)] = DatasetAnalyzer.analyze(v)
        return FeedDictDatasetType(types)

    def serialize(self, instance: dict):
        self._check_type(instance, dict, SerializationError)
        try:
            items = {self.get_key(k): v for k, v in instance.items()}
        except ValueError as e:
            raise SerializationError(e)
        return super().serialize(self, items)

    @staticmethod
    def get_key(k):
        if isinstance(k, tf.Tensor):
            return k.name
        elif isinstance(k, str):
            return k
        else:
            raise ValueError(f'Unknown key type {type(k).__name__} for key {k} in feed_dict')

    @DictDatasetType.requirements.getter
    def requirements(self) -> Requirements:
        return Requirements([InstallableRequirement.from_module(tf)]) + super().requirements.fget(self)


class FeedDictHook(CanIsAMustHookMixin, DatasetHook):
    """
    DatasetHook for tensorflow feed dict
    """

    def must_process(self, obj) -> bool:
        """
        :param obj: obj to check
        :return: `True` if obj is mapping and any of it's keys are tf.Tensor instance
        """
        is_mapping = isinstance(obj, Mapping)
        return is_mapping and any(isinstance(k, tf.Tensor) for k in obj.keys())

    def process(self, obj, **kwargs) -> DatasetType:
        """
        :param obj: obj to process
        :return: :class:`FeedDictDatasetType` instance
        """
        return FeedDictDatasetType.from_feed_dict(obj)



import pickle
from abc import abstractmethod
from typing import Tuple

from pyjackson.decorators import type_field

from algolink.core.errors import ArtifactExistsError, DatasetExistsError, NoSuchArtifactError, NoSuchDataset
from algolink.core.objects import ArtifactCollection, DatasetType
from algolink.core.objects.artifacts import Blobs, InMemoryBlob
from algolink.core.objects.base import AlgoLinkParams
from algolink.core.objects.dataset_source import Dataset, DatasetSource
from algolink.repository import ArtifactRepository
from algolink.repository.dataset.base import DatasetRepository


@type_field('type')
class DatasetReader(AlgoLinkParams):
    """ABC for reading Dataset from files (artifacts) to use with ArtifactDatasetSource"""

    @abstractmethod
    def read(self, artifacts: ArtifactCollection) -> Dataset:
        """Method to read Dataset from artifacts

        :param artifacts: artifacts to read"""


@type_field('type')
class DatasetWriter(AlgoLinkParams):
    """ABC for writing Dataset to files (artifacts) to use with ArtifactDatasetSource"""

    @abstractmethod
    def write(self, dataset: Dataset) -> Tuple[DatasetReader, ArtifactCollection]:
        """Method to write dataset to artifacts

        :param dataset: dataset to write
        :returns: tuple of DatasetReader and ArtifactCollection.
        DatasetReader must produce the same dataset if used with same artifacts"""


class OneFileDatasetReader(DatasetReader):
    def __init__(self, dataset_type: DatasetType):
        self.dataset_type = dataset_type

    def read(self, artifacts: ArtifactCollection) -> Dataset:
        payload = artifacts.bytes_dict()[OneFileDatasetWriter.FILENAME]
        return Dataset(self.convert(payload), self.dataset_type)

    @abstractmethod
    def convert(self, payload: bytes):
        """"""  # TODO


class OneFileDatasetWriter(DatasetWriter):
    FILENAME = 'data'

    def convert(self, instance) -> bytes:
        """"""

    def write(self, dataset: Dataset) -> Tuple[DatasetReader, ArtifactCollection]:
        return OneFileDatasetReader(dataset.dataset_type), \
               Blobs.from_blobs({self.FILENAME: InMemoryBlob(self.convert(dataset.data))})


class PrimitiveDatasetReader(OneFileDatasetReader):
    def convert(self, payload: bytes):
        return self.dataset_type.to_type(payload)


class PrimitiveDatasetWriter(OneFileDatasetWriter):
    def convert(self, instance) -> bytes:
        return str(instance).encode('utf8')


class PickleReader(OneFileDatasetReader):
    def convert(self, payload: bytes):
        return pickle.loads(payload)


class PickleWriter(OneFileDatasetWriter):
    def convert(self, instance) -> bytes:
        return pickle.dumps(instance)


class ArtifactDatasetRepository(DatasetRepository):
    """DatasetRpository implementation that saves datasets as artifacts to ArtifactRepository

    :param repo: underlying ArtifactRepository"""

    ARTIFACT_TYPE = 'datasets'

    def __init__(self, repo: ArtifactRepository):
        self.repo = repo

    def save(self, dataset_id: str, dataset: Dataset) -> DatasetSource:
        writer = dataset.get_writer()
        if writer is None:
            raise ValueError(f'{dataset.dataset_type} does not support artifact persistance')

        reader, artifacts = writer.write(dataset)
        with artifacts.blob_dict() as blobs:
            try:
                pushed = self.repo.push_artifact(self.ARTIFACT_TYPE, dataset_id, blobs)
            except ArtifactExistsError as e:
                raise DatasetExistsError(dataset_id, self, e)
        return ArtifactDatasetSource(reader, pushed, dataset.dataset_type)

    def delete(self, dataset_id: str):
        try:
            self.repo.delete_artifact(self.ARTIFACT_TYPE, dataset_id)
        except NoSuchArtifactError as e:
            raise NoSuchDataset(dataset_id, self, e)


class ArtifactDatasetSource(DatasetSource):
    """DatasetSource for reading datasets from ArtifactDatasetRepository

    :param reader: DatasetReader for this dataset
    :param artifacts: ArtifactCollection with actual files
    :param dataset_type: DatasetType of contained dataset"""

    def __init__(self, reader: DatasetReader, artifacts: ArtifactCollection, dataset_type: DatasetType):
        super(ArtifactDatasetSource, self).__init__(dataset_type)
        self.reader = reader
        self.artifacts = artifacts

    def read(self) -> Dataset:
        return self.reader.read(self.artifacts)



from typing import List

import algolink.client.base
from algolink import AlgoLink
from .expose import ExposedMethod, get_exposed_method
from algolink.core.objects import Image, Model, Pipeline, Project, RuntimeEnvironment, RuntimeInstance, Task
from algolink.repository.metadata.base import MetadataRepository

AUTOGEN_CLASSES = {
    MetadataRepository: '#  ########## AUTOGEN META END #',
    Project: '#  ########## AUTOGEN PROJECT END #',
    Task: '#  ########## AUTOGEN TASK END #',
    Model: '#  ########## AUTOGEN MODEL END #',
    Pipeline: '#  ########## AUTOGEN PIPELINE END #',
    Image: '#  ########## AUTOGEN IMAGE END #',
    RuntimeInstance: '#  ########## AUTOGEN INSTANCE END #',
    RuntimeEnvironment: '#  ########## AUTOGEN ENVIRONMENT END #'
}

FILENAME = algolink.client.base.__file__


def find_exposed_methods(base_class, new_only=True) -> List[ExposedMethod]:
    new_methods = []
    for name in dir(base_class):
        if name.startswith('_'):
            continue
        value = getattr(base_class, name)
        exposed = get_exposed_method(value)
        if exposed is not None and (not new_only or exposed.name not in AlgoLink.__dict__):
            new_methods.append(exposed)
    return new_methods


def patch(classes, filename, dry_run=True):
    anchors = {a: cls for cls, a in AUTOGEN_CLASSES.items() if cls in classes}
    code = []
    with open(filename, 'r') as f:
        for line in f:
            strip = line.strip()
            if strip in anchors:
                base_class = anchors[strip]
                for exposed in find_exposed_methods(base_class):
                    code.append(exposed.generate_code() + '\n\n')
            code.append(line)

    if dry_run:
        print(''.join(code))
    else:
        with open(filename, 'w') as f:
            f.write(''.join(code))


def clear(filename, dry_run=True):
    code = []
    is_autogen = False
    with open(filename, 'r') as f:
        for line in f:
            sline = line.strip()
            if sline.startswith('#') and 'AUTOGEN' in sline:
                if 'END' in sline:
                    is_autogen = False
                else:
                    is_autogen = True
                    code.append(line + '\n')
            if not is_autogen:
                code.append(line)

    if dry_run:
        print(''.join(code))
    else:
        with open(filename, 'w') as f:
            f.write(''.join(code))


def main():
    dry_run = True
    # clear(FILENAME, dry_run)
    patch(AUTOGEN_CLASSES, FILENAME, dry_run)


if __name__ == '__main__':
    main()



import uuid

import yaml
from aiohttp import web
from aiohttp_swagger import setup_swagger

from algolink.runtime.interface import Interface
from algolink.runtime.interface.base import InterfaceDescriptor
from algolink.runtime.openapi.spec import create_spec
from algolink.runtime.server import BaseHTTPServer, HTTPServerConfig, MalformedHTTPRequestException
from algolink.utils.log import rlogger


def create_executor_function(interface: Interface, method: str, spec: dict):
    """
    Creates a view function for specific interface method

    :param interface: :class:`.Interface` instance
    :param method: method name
    :param spec: openapi spec for this instance
    :return: callable view function
    """

    async def ef(request):
        algolink_id = str(uuid.uuid4())
        rlogger.debug('Headers for [%s]: %s', request.headers)

        try:
            if request.content_type == 'application/json':
                request_data = BaseHTTPServer._deserialize_json(interface, method, await request.json())
            else:
                request_data = {k: v.file for k, v in dict(await request.post()).items()}

            result = BaseHTTPServer._execute_method(interface, method, request_data, algolink_id)

            if isinstance(result, bytes):
                return web.Response(body=result, content_type='image/png')
            return web.json_response(result)
        except MalformedHTTPRequestException as e:
            return web.json_response(e.response_body(), status=e.code())

    ef.__doc__ = f"\n---\n{yaml.dump(spec)}\n"

    return ef


def create_interface_routes(app, interface: Interface):
    for method in interface.exposed_methods():
        sig = interface.exposed_method_signature(method)
        rlogger.debug('registering %s with input type %s and output type %s', method, sig.args, sig.output)

        spec = create_spec(method, sig, str(Interface), interface.exposed_method_docs(method))
        executor_function = create_executor_function(interface, method, spec)
        app.router.add_post('/' + method, executor_function)


def create_schema_route(app, interface: Interface):
    schema = InterfaceDescriptor.from_interface(interface).to_dict()
    rlogger.debug('Creating /interface.json route with schema: %s', schema)
    app.router.add_get('/interface.json', lambda request: web.json_response(schema))


def create_misc_routes(app):
    async def redirect_to_swagger(request):
        raise web.HTTPFound('/apidocs')

    app.router.add_get('/health', lambda request: web.Response(body='OK'))
    app.router.add_get('/', redirect_to_swagger)


class AIOHTTPServer(BaseHTTPServer):
    def __init__(self):
        # we do not reference real aiohttp objects here and this breaks `get_object_requirements`
        import aiohttp_swagger
        self.__requires = aiohttp_swagger
        super().__init__()

    def run(self, interface: Interface):
        app = web.Application()
        create_interface_routes(app, interface)
        create_schema_route(app, interface)
        create_misc_routes(app)
        setup_swagger(app, swagger_url="/apidocs", ui_version=3)

        rlogger.debug('Running aiohttp on %s:%s', HTTPServerConfig.host, HTTPServerConfig.port)
        web.run_app(app, host=HTTPServerConfig.host, port=HTTPServerConfig.port)



from .dataset import LightGBMDatasetHook, LightGBMDatasetType
from .model import LightGBMModelHook, LightGBMModelWrapper
from .requirement import LightGBMRequirementHook

__all__ = ['LightGBMModelWrapper', 'LightGBMModelHook', 'LightGBMDatasetHook', 'LightGBMDatasetType',
           'LightGBMRequirementHook']



import os
import shutil
from typing import Dict, List, Optional, TypeVar, Union

from pyjackson import read, write
from pyjackson.utils import resolve_subtype

from algolink.core.errors import ExistingImageError, ExistingInstanceError
from algolink.core.objects import Image, Model, Pipeline, RuntimeEnvironment, RuntimeInstance, Task
from algolink.core.objects.core import AlgoLinkObject, Project, WithDatasetRepository
from algolink.core.objects.dataset_source import Dataset
from algolink.repository import DatasetRepository
from algolink.repository.artifact import ArtifactRepository
from algolink.repository.artifact.inmemory import InMemoryArtifactRepository
from algolink.repository.artifact.local import LocalArtifactRepository
from algolink.repository.dataset.artifact import ArtifactDatasetRepository
from algolink.repository.metadata import MetadataRepository
from algolink.repository.metadata.base import ProjectVar, TaskVar
from algolink.repository.metadata.local import LocalMetadataRepository
from algolink.runtime.server import Server
from algolink.utils.importing import module_importable

T = TypeVar('T', bound=AlgoLinkObject)


class AlgoLink:
    """Main entry point for algolink

    This is the client for AlgoLink API. It can save, load and build Models, Tasks and Projects.
    AlgoLink instance can be obtained from factory methods like :meth:`~algolink.AlgoLink.local` for local client,
    :meth:`~algolink.AlgoLink.inmemory` for inmemory client.

    You can save client config with :meth:`~algolink.AlgoLink.save_client_config` and
    later restore it with :meth:`~algolink.AlgoLink.from_config_file`

    :param meta_repo: :class:`~algolink.repository.MetadataRepository` instance to save metadata
    :param artifact_repo: :class:`~algolink.repository.ArtifactRepository` instance to save artifacts
    """

    default_server: Server = None
    default_env: RuntimeEnvironment = None

    def __init__(self, meta_repo: MetadataRepository, artifact_repo: ArtifactRepository,
                 dataset_repo: DatasetRepository = None):
        self.meta_repo = meta_repo
        self.artifact_repo = artifact_repo
        self.dataset_repo = dataset_repo or ArtifactDatasetRepository(self.artifact_repo)

    def _bind(self, obj: Optional[Union[T, List[T]]]) -> Optional[Union[T, List[T]]]:
        """Binds AlgoLinkObject to meta and art repo of this client instance

        :param obj: subclass of AlgoLinkObject instance or list of them
        """
        if obj is None:
            return
        if isinstance(obj, list):
            for o in obj:
                self._bind(o)
        else:

            obj.bind_meta_repo(self.meta_repo).bind_artifact_repo(self.artifact_repo)
            if isinstance(obj, WithDatasetRepository):
                obj.bind_dataset_repo(self.dataset_repo)
        return obj

    def push_model(self, model: Model, task: Task = None) -> Model:
        """
        Pushes :py:class:`~algolink.core.objects.Model` instance into metadata and artifact repositories

        :param model: :py:class:`~algolink.core.objects.Model` instance
        :param task: :py:class:`~algolink.core.objects.Task` instance to save model to. Optional if model already has
        task
        :return: same saved :py:class:`~algolink.core.objects.Model` instance
        """
        self._bind(model)
        return model.push(task)

    def create_model(self, model_object, model_input, model_name: str = None, *,
                     project_name: str = 'default_project', task_name: str = 'default_task', **kwargs):
        """
        This function creates algolink model.
        Creates model, task and project (if needed) and pushes it to repo

        :param model_object: object containing model.
        :param model_input: model input.
        :param model_name: model name to create.
        :param project_name: project name.
        :param task_name: task name.
        :param kwargs: other arguments for model

        :return: :class:`~algolink.core.objects.Model` instance representing
        """
        task = self.get_or_create_task(project_name, task_name)
        return task.create_and_push_model(model_object, model_input, model_name, **kwargs)

    def get_model(self, model_name: str, task: TaskVar, project: ProjectVar = None,
                  load_artifacts: bool = True) -> Model:
        """
        Load model from repository

        :param model_name: model name to load
        :param task: :py:class:`~algolink.core.objects.Task` instance or task name to load model from
        :param project: :py:class:`~algolink.core.objects.Project` instance or project name to load task from
        :param load_artifacts: if True, load model artifact into wrapper
        :return: :py:class:`~algolink.core.objects.Model` instance
        """
        model: Model = self.meta_repo.get_model_by_name(model_name, task, project)
        if model is not None and load_artifacts:
            model.load()
        return self._bind(model)

    def create_image(self, obj, name: str = None, task: Task = None, server: Server = None,
                     environment: RuntimeEnvironment = None,
                     debug=False, skip_build=False, builder_args: Dict[str, object] = None, **kwargs) -> Image:
        """
        Builds image of model service and stores it to repository

        :param obj: model/list of models/pipeline or any object that has existing Hook for it to wrap into service
        :param name: name of image to build
        :param task: task to put image into
        :param server: server to build image with
        :param environment: env to build for
        :param debug: flag to build debug image
        :param skip_build: wheter to skip actual image build
        :param builder_args: kwargs for image.build
        :param kwargs: additional kwargs for builder
        :return: :class:`~algolink.core.objects.Image` instance representing built image
        """
        from algolink.core.analyzer.buildable import BuildableAnalyzer

        if server is None:
            server = self.get_default_server()

        if environment is None:
            environment = self.get_default_environment()

        buildable = BuildableAnalyzer.analyze(obj, server=server, debug=debug).bind_meta_repo(self.meta_repo)
        task = task or buildable.task
        if task is None:
            raise ValueError(f'cannot infer task for buildable {buildable}, please provide it manually')
        if name is not None and self.meta_repo.get_image_by_name(name, task) is not None:
            raise ExistingImageError(name)

        builder = environment.params.get_builder()
        image = Image(name, buildable)
        image.params = builder.create_image(image.name, environment.params, **kwargs)
        image.task = task
        image.environment = environment
        self.meta_repo.create_image(image)
        if not skip_build:
            try:
                image.build(**(builder_args or {}))
            except Exception:
                self.meta_repo.delete_image(image)
                raise
        return self._bind(self.meta_repo.save_image(image))

    def create_instance(self, image: Image, name: str = None, environment: RuntimeEnvironment = None, run=False,
                        runner_kwargs: Dict[str, object] = None,
                        **instance_kwargs) -> RuntimeInstance:
        """
        Runs model service instance and stores it to repository

        :param image: image to run instance from
        :param name: name of instance to run
        :param environment: environment to run instance in, if no given `localhost` is used
        :param run:  whether to automatically run instance after creation
        :param runner_kwargs: additional parameters for runner
        :param instance_kwargs: additional parameters for instance
        :return: :class:`~algolink.core.objects.RuntimeInstance` instance representing run instance
        """

        if environment is None:
            environment = self.get_default_environment()

        if name is not None and self.meta_repo.get_instance_by_name(name, image, environment) is not None:
            raise ExistingInstanceError(name)

        runner = environment.params.get_runner()

        instance = RuntimeInstance(name)

        instance.params = runner.create_instance(instance.name, **instance_kwargs)
        instance.image = image
        instance.environment = environment
        instance.bind_runner(runner)
        instance = self.meta_repo.create_instance(instance)
        if run:
            runner_kwargs = runner_kwargs or {}
            instance.run(**runner_kwargs)
        return self._bind(instance)

    def build_and_run_instance(self, obj, name: str = None, task: Task = None, environment: RuntimeEnvironment = None,
                               builder_kwargs: Dict[str, object] = None, runner_kwargs: Dict[str, object] = None,
                               instance_kwargs: Dict[str, object] = None) -> RuntimeInstance:
        """
        Builds image of model service, immediately runs service and stores both image and instance to repository

        :param obj: buildable object to wrap into service
        :param name: name of image and instance to be built and run respectively
        :param task: task to put image into
        :param environment: environment to run instance in, if no given `localhost` is used
        :param builder_kwargs: additional kwargs for builder
        :param runner_kwargs: additional parameters for runner. Full list can be seen in
                    https://docker-py.readthedocs.io/en/stable/containers.html
        :param instance_kwargs: additional parameters for instance
        :return: :class:`~algolink.core.objects.RuntimeInstance` instance representing run instance
        """
        instance_kwargs = instance_kwargs or {}
        runner_kwargs = runner_kwargs or {}
        builder_kwargs = builder_kwargs or {}
        image = self.create_image(obj, name, task, environment=environment, builder_args=builder_kwargs)
        return self.create_instance(image, name, environment, **instance_kwargs).run(**runner_kwargs)

    @classmethod
    def local(cls, path=None, clear=False) -> 'AlgoLink':
        """
        Get an instance of :class:`~algolink.AlgoLink` that stores metadata and artifacts on local filesystem

        :param path: path to storage dir. If None, `.algolink` dir is used
        :param clear: if True, erase previous data from storage
        """
        path = path or '.algolink'
        if clear and os.path.exists(path):
            shutil.rmtree(path)
        meta_repo = LocalMetadataRepository(os.path.join(path, 'metadata.json'))
        artifact_repo = LocalArtifactRepository(os.path.join(path, 'artifacts'))
        return AlgoLink(meta_repo, artifact_repo)

    @classmethod
    def inmemory(cls) -> 'AlgoLink':
        """
        Get an instance of :class:`~algolink.AlgoLink` with inmemory repositories
        """
        return AlgoLink(LocalMetadataRepository(), InMemoryArtifactRepository())

    @classmethod
    def custom_client(cls, metadata: Union[str, MetadataRepository], artifact: Union[str, ArtifactRepository],
                      meta_kwargs: dict = None, artifact_kwargs: dict = None) -> 'AlgoLink':
        """
        Create custom AlgoLink client from metadata and artifact repositories.

        :param metadata: :class:`~algolink.repository.MetadataRepository` instance or pyjackson subtype type name
        :param artifact: :class:`~algolink.repository.ArtifactRepository` instance or pyjackson subtype type name
        :param meta_kwargs: kwargs for metadata repo __init__ if subtype type name was provided
        :param artifact_kwargs: kwargs for artifact repo __init__ if subtype type name was provided
        :return: :class:`~algolink.AlgoLink` instance
        """
        if isinstance(metadata, str):
            metadata_type = resolve_subtype(MetadataRepository, {'type': metadata})
            meta_kwargs = meta_kwargs or {}
            metadata = metadata_type(**meta_kwargs)

        if isinstance(artifact, str):
            artifact_type = resolve_subtype(ArtifactRepository, {'type': artifact})
            artifact_kwargs = artifact_kwargs or {}
            artifact = artifact_type(**artifact_kwargs)

        return AlgoLink(metadata, artifact)

    @classmethod
    def from_config_file(cls, filepath) -> 'AlgoLink':
        """
        Read and create AlgoLink instance from config file

        :param filepath: path to read config from
        :return: :class:`~algolink.AlgoLink` instance
        """
        return read(filepath, AlgoLink)

    def save_client_config(self, filepath):
        """
        Save current client config to a file

        :param filepath: path to file
        """
        write(filepath, self, AlgoLink)

    def get_default_server(self):
        """
        :return: Default server implementation for this client
        """
        if self.default_server is None:
            from algolink.ext.flask import FlaskServer
            self.default_server = FlaskServer()
        return self.default_server

    def get_default_environment(self):
        """
        Creates (if needed) and returns default runtime environment

        :return: saved instance of :class:`.RuntimeEnvironment`
        """
        if self.default_env is not None:
            return self.default_env
        env_name = 'docker_localhost'
        self.default_env = self.get_environment(env_name)
        if self.default_env is None:
            if not module_importable('docker'):
                raise RuntimeError("Can't build docker container: docker module is not installed. Install it "
                                   "with 'pip install docker'")

            from algolink.ext.docker import DockerEnv
            self.default_env = RuntimeEnvironment(env_name, params=DockerEnv())
            self.default_env = self.push_environment(self.default_env)
        return self.default_env

    #  ########## AUTOGEN #####

    #  ########## AUTOGEN META #####

    def push_environment(self, environment: 'RuntimeEnvironment') -> RuntimeEnvironment:
        """
        Creates runtime environment in the repository

        :param environment: runtime environment to create
        :return: created runtime environment
        :exception: :exc:`.errors.ExistingEnvironmentError` if given runtime environment has the same name as existing
        """
        return self._bind(self.meta_repo.create_environment(environment))

    def get_environment(self, name: str) -> Optional[RuntimeEnvironment]:
        """
        Finds runtime environment by name.

        :param name: expected runtime environment name
        :return: found runtime environment if exists or `None`
        """
        return self._bind(self.meta_repo.get_environment_by_name(name))

    def get_environments(self) -> List[RuntimeEnvironment]:
        """
        Gets a list of runtime environments

        :return: found runtime environments
        """
        return self._bind(self.meta_repo.get_environments())

    def get_image(self, image_name: str, task: TaskVar, project: ProjectVar = None) -> Optional['Image']:
        """
        Finds image by name in given model, task and project.

        :param image_name: expected image name
        :param task: task to search for image in
        :param project: project to search for image in
        :return: found image if exists or `None`
        """
        return self._bind(self.meta_repo.get_image_by_name(image_name, task, project))

    def get_images(self, task: TaskVar, project: ProjectVar = None) -> List['Image']:
        """
        Gets a list of images in given model, task and project

        :param task: task to search for images in
        :param project: project to search for images in
        :return: found images
        """
        return self._bind(self.meta_repo.get_images(task, project))

    def get_instance(self, instance_name: str, image: Union[int, 'Image'],
                     environment: Union[int, 'RuntimeEnvironment']) -> Optional['RuntimeInstance']:
        """
        Finds instance by name in given image and environment.

        :param instance_name: expected instance name
        :param image: image (or id) to search for instance in
        :param environment: environment (or id) to search for instance in
        :return: found instance if exists or `None`
        """
        return self._bind(self.meta_repo.get_instance_by_name(instance_name, image, environment))

    def get_instances(self, image: Union[int, 'Image'] = None,
                      environment: Union[int, 'RuntimeEnvironment'] = None) -> List['RuntimeInstance']:
        """
        Gets a list of instances in given image or environment

        :param image: image (or id) to search for instances in
        :param environment: environment (or id) to search for instances in
        :return: found instances
        """
        return self._bind(self.meta_repo.get_instances(image, environment))

    def get_models(self, task: TaskVar, project: ProjectVar = None) -> List['Model']:
        """
        Gets a list of models in given project and task

        :param task: task to search for models in
        :param project: project to search for models in
        :return: found models
        """
        return self._bind(self.meta_repo.get_models(task, project))

    def get_or_create_project(self, name: str) -> Project:
        """
        Creates a project if not exists or gets existing project otherwise.

        :param name: project name
        :return: project
        """
        return self._bind(self.meta_repo.get_or_create_project(name))
    ## create experiment -leepand
    def get_or_create_experiment(self, name: str) -> Experiment:
        """
        Creates a project if not exists or gets existing project otherwise.

        :param name: project name
        :return: project
        """
        return self._bind(self.meta_repo.get_or_create_experiment(name))        
    def get_or_create_task(self, project: str, task_name: str) -> Task:
        """
        Creates a task if not exists or gets existing task otherwise.

        :param project: project to search/create task in
        :param task_name: expected name of task
        :return: created/found task
        """
        return self._bind(self.meta_repo.get_or_create_task(project, task_name))
    # model tracking --leepand
    def get_or_create_experiment(self, experiment: str, submodel_name: str) -> SubModel:
        """
        Creates a experiment if not exists or gets existing submodel otherwise.

        :param project: experiment to search/create submodel in
        :param submodel_name: expected name of submodel
        :return: created/found submodel
        """
        return self._bind(self.meta_repo.get_or_create_submodel(experiment, submodel_name))

    def get_pipeline(self, pipeline_name: str, task: TaskVar,
                     project: ProjectVar = None) -> Optional['Pipeline']:
        """
        Finds model by name in given task and project.

        :param pipeline_name: expected pipeline name
        :param task: task to search for pipeline in
        :param project: project to search for pipeline in
        :return: found pipeline if exists or `None`
        """
        return self._bind(self.meta_repo.get_pipeline_by_name(pipeline_name, task, project))

    def get_pipelines(self, task: TaskVar, project: ProjectVar = None) -> List['Pipeline']:
        """
        Gets a list of pipelines in given project and task

        :param task: task to search for models in
        :param project: project to search for models in
        :return: found pipelines
        """
        return self._bind(self.meta_repo.get_pipelines(task, project))

    def get_project(self, name: str) -> Optional['Project']:
        """
        Finds project in the repository by name

        :param name: name of the project to return
        :return: found project if exists or `None`
        """
        return self._bind(self.meta_repo.get_project_by_name(name))

    def get_projects(self) -> List['Project']:
        """
        Gets all projects in the repository

        :return: all projects in the repository
        """
        return self._bind(self.meta_repo.get_projects())

    def get_task(self, project: ProjectVar, task_name: str) -> Optional['Task']:
        """
        Finds task with given name in given project

        :param project: project to search for task in
        :param task_name: expected name of task
        :return: task if exists or `None`
        """
        return self._bind(self.meta_repo.get_task_by_name(project, task_name))

    def get_tasks(self, project: ProjectVar) -> List['Task']:
        """
        Gets a list of tasks for given project

        :param project: project to search for tasks in
        :return: project tasks
        """
        return self._bind(self.meta_repo.get_tasks(project))

    #  ########## AUTOGEN META END #

    #  ########## AUTOGEN PROJECT #

    def delete_project(self, project: Project, cascade: bool = False):
        """"
        Deletes project and(if required) all tasks associated with it from metadata repository

        :param project: project to delete
        :param cascade: whether should project be deleted with all associated tasks
        :return: Nothing
        """
        return project.delete(cascade)

    #  ########## AUTOGEN PROJECT END #

    #  ########## AUTOGEN TASK #

    def delete_task(self, task: Task, cascade: bool = False):
        """"
        Deletes task from metadata

        :param task: task to delete
        :param cascade: whether should task be deleted with all associated objects
        :return: Nothing
        """
        return task.delete(cascade)

    #  ########## AUTOGEN TASK END #

    #  ########## AUTOGEN MODEL #

    def delete_model(self, model: Model, force: bool = False):
        """"
        Deletes model from metadata and artifact repositories

        :param model: model to delete
        :param force: whether model artifacts' deletion errors should be ignored, default is false
        :return: Nothing
        """
        return model.delete(force)

    #  ########## AUTOGEN MODEL END #

    #  ########## AUTOGEN PIPELINE #

    def delete_pipeline(self, pipeline: Pipeline):
        """"Deletes pipeline from metadata

        :param pipeline: pipeline to delete
        """
        return pipeline.delete()

    #  ########## AUTOGEN PIPELINE END #

    #  ########## AUTOGEN IMAGE #

    def delete_image(self, image: Image, meta_only: bool = False, cascade: bool = False):
        """"
        Deletes existing image from metadata repository and image provider

        :param image: image ot delete
        :param meta_only: should image be deleted only from metadata
        :param cascade: whether to delete nested RuntimeInstances
        """
        return image.delete(meta_only, cascade)

    #  ########## AUTOGEN IMAGE END #

    #  ########## AUTOGEN INSTANCE #

    def delete_instance(self, instance: RuntimeInstance, meta_only: bool = False):
        """"
        Stops instance of model service and deletes it from repository

        :param instance: instance to delete
        :param meta_only: only remove from metadata, do not stop instance
        :return: nothing
        """
        return instance.delete(meta_only)

    #  ########## AUTOGEN INSTANCE END #

    #  ########## AUTOGEN ENVIRONMENT #

    def delete_environment(self, environment: RuntimeEnvironment, meta_only: bool = False, cascade: bool = False):
        """"
        Deletes environment from metadata repository and(if required) stops associated instances

        :param environment: environment to delete
        :param meta_only: wheter to only delete metadata
        :param cascade: Whether should environment be deleted with all associated instances
        :return: Nothing
        """
        return environment.delete(meta_only, cascade)

    #  ########## AUTOGEN ENVIRONMENT END #

    #  ########## AUTOGEN END #

    def create_dataset(self, data, target=None):
        # TODO persisting to art repo?
        return Dataset.from_object(data)

    def create_metric(self, metric_obj):
        from algolink.core.analyzer.metric import MetricAnalyzer
        return MetricAnalyzer.analyze(metric_obj)



import sys
from importlib import _bootstrap


"""
Copied from django.utils.module_loading
"""


def import_module(name, package=None):
    """Import a module.

    The 'package' argument is required when performing a relative import. It
    specifies the package to use as the anchor point from which to resolve the
    relative import to an absolute import.

    """
    level = 0
    if name.startswith('.'):
        if not package:
            msg = ("the 'package' argument is required to perform a relative "
                   "import for {!r}")
            raise TypeError(msg.format(name))
        for character in name:
            if character != '.':
                break
            level += 1
    return _bootstrap._gcd_import(name[level:], package, level)


def import_string(dotted_path):
    """
    Import a dotted module path and return the attribute/class designated by the
    last name in the path. Raise ImportError if the import failed.
    """
    try:
        module_path, class_name = dotted_path.rsplit('.', 1)
    except ValueError as err:
        raise ImportError("%s doesn't look like a module path" % dotted_path) from err

    module = import_module(module_path)

    try:
        return getattr(module, class_name)
    except AttributeError as err:
        raise ImportError('Module "%s" does not define a "%s" attribute/class' % (
            module_path, class_name)) from err


def module_importable(module_name):
    try:
        import_module(module_name)
        return True
    except ImportError:
        return False


def module_imported(module_name):
    """
    Checks if module already imported

    :param module_name: module name to check
    :return: `True` or `False`
    """
    return sys.modules.get(module_name) is not None



from abc import abstractmethod

from algolink.core.analyzer import Hook, analyzer_class
from .base import LibHookMixin
from ..objects.metric import CallableMetric, CallableMetricWrapper, LibFunctionMetric, Metric


class MetricHook(Hook):
    """
    Base hook type for :class:`DatasetAnalyzer`.
    Analysis result is an instance of :class:`~algolink.core.objects.DatasetType`
    """

    @abstractmethod
    def process(self, obj, **kwargs) -> Metric:
        """
        Analyzes obj and returns result. Result type is determined by specific Hook class sub-hierarchy

        :param obj: object to analyze
        :param kwargs: additional information to be used for analysis
        :return: analysis result
        """
        pass  # pragma: no cover


MetricAnalyzer = analyzer_class(MetricHook, Metric)


class LibFunctionMixin(MetricHook, LibHookMixin):
    invert = False
    default_args = {}

    def get_args(self, obj):
        return self.default_args

    def process(self, obj, **kwargs) -> Metric:
        return LibFunctionMetric(f'{obj.__module__}.{obj.__name__}', self.get_args(obj), invert_input=self.invert)


class CallableMetricHook(MetricHook):
    def process(self, obj, **kwargs) -> Metric:
        return CallableMetric(CallableMetricWrapper.from_callable(obj))

    def can_process(self, obj) -> bool:
        return callable(obj)

    def must_process(self, obj) -> bool:
        return False



from typing import Any, Dict

from algolink.core.analyzer.buildable import BuildableAnalyzer
from algolink.core.objects import Image, RuntimeEnvironment, RuntimeInstance
from .base import DockerEnv
from .builder import DockerBuilder
from .runner import DockerRunner
from algolink.runtime.server import Server


def build_docker_image(name: str, obj, server: Server = None, env: DockerEnv = None, tag: str = 'latest',
                       repository: str = None, force_overwrite: bool = False, **kwargs) -> Image:
    """Build docker image from object

    :param name: name of the resultimg image
    :param obj: obj to build image. must be convertible to Buildable: Model, Pipeline, list of one of those, etc.
    :param server: server to build image with
    :param env: DockerEnv to build in. Default - local docker daemon
    :param tag: image tag
    :param repository: image repository
    :param force_overwrite: wheter to force overwrite existing image
    :parma kwargs: additional arguments for DockerBuilder.build_image
    """
    if server is None:
        try:
            from algolink.ext.flask import FlaskServer
            server = FlaskServer()
        except ImportError:
            raise RuntimeError('cannot use default FlaskServer - flask or flasgger are not installed')
    env = env or DockerEnv()
    source = BuildableAnalyzer.analyze(obj, server=server)
    builder: DockerBuilder = env.get_builder()
    params = builder.create_image(name, env, tag, repository)
    builder.build_image(source, params, env, force_overwrite, **kwargs)
    image = Image(name, source, params=params)
    image.environment = RuntimeEnvironment('temp_env', params=env)
    return image.bind_builder(builder)


def run_docker_instance(image: Image, name: str = None, env: DockerEnv = None,
                        port_mapping: Dict[int, int] = None, instance_kwargs: Dict[str, Any] = None, rm: bool = False,
                        detach: bool = True, **kwargs) -> RuntimeInstance:
    """Create and run docker container

    :param image: image to build from
    :param name: name of the container. defaults to image name
    :param env: DockerEnv to run in. Default - local docker daemon
    :param port_mapping: port mapping for container
    :param instance_kwargs: additional DockerInstance args
    :param rm: wheter to remove container on exit
    :param detach: wheter to detach from container after run
    :param kwargs: additional args for DockerRunner.run
    """
    env = env or DockerEnv()
    name = name or image.name
    runner: DockerRunner = env.get_runner()
    params = runner.create_instance(name, port_mapping, **instance_kwargs or {})
    runner.run(params, image.params, env, rm, detach, **kwargs)
    instance = RuntimeInstance(name, params=params, image_id=image.id).bind_runner(runner)
    instance.environment = RuntimeEnvironment('temp_env', params=env)
    return instance



import contextlib
import os
import pickle
import typing
from abc import abstractmethod
from functools import wraps
from importlib import import_module
from io import BytesIO
from pickle import _Unpickler
from uuid import uuid4

from pyjackson import dumps, read
from pyjackson.core import Unserializable
from pyjackson.decorators import type_field
from pyjackson.utils import get_class_fields

from algolink.core.analyzer.dataset import DatasetAnalyzer
from .artifacts import ArtifactCollection, Blob, Blobs, CompositeArtifactCollection, InMemoryBlob
from .base import AlgoLinkParams
from .dataset_type import DatasetType
from .requirements import InstallableRequirement, Requirements
from algolink.utils.fs import switch_curdir
from algolink.utils.module import get_object_requirements
from algolink.utils.pickling import AlgoLinkPickler

FilesContextManager = typing.ContextManager[ArtifactCollection]
MethodArg = DatasetType
MethodReturn = DatasetType
Method = typing.Tuple[str, MethodArg, MethodReturn]
Methods = typing.Dict[str, Method]


@type_field('type')
class ModelIO(AlgoLinkParams):
    """
    Helps model wrapper with IO

    Must be pyjackson-serializable
    """

    @abstractmethod
    def dump(self, model) -> FilesContextManager:
        """
        Must return context manager with :class:`~algolink.core.objects.ArtifactCollection`
        :return: :class:`~algolink.core.objects.ArtifactCollection`
        """
        pass  # pragma: no cover

    @abstractmethod
    def load(self, path):
        """
        Must load and return model
        :param path: path to load model from
        :return: model object
        """
        pass  # pragma: no cover


@type_field('type')
class ModelWrapper(AlgoLinkParams):
    """
    Base class for model wrapper. Wrapper is an object that can save, load and inference a model

    Must be pyjackson-serializable
    """
    type = None
    methods_json = 'methods.json'
    requirements_json = 'requirements.json'

    def __init__(self, io: ModelIO):
        self.model = None
        self.methods: typing.Optional[Methods] = None
        self.requirements: Requirements = None
        self.io = io
        self.curdir = '.'

    @contextlib.contextmanager
    def dump(self) -> FilesContextManager:
        with self.io.dump(self.model) as artifact:
            yield artifact + Blobs({
                self.methods_json: InMemoryBlob(dumps(self.methods).encode('utf-8')),
                self.requirements_json: InMemoryBlob(dumps(self.requirements).encode('utf-8'))
            })

    def load(self, path):
        self.curdir = os.path.abspath(path)
        self.model = self.io.load(path)
        self.methods = read(os.path.join(path, self.methods_json), typing.Optional[Methods])
        self.requirements = read(os.path.join(path, self.requirements_json), Requirements)

    def bind_model(self, model, input_data=None, **kwargs):
        """
        Bind model object to this wrapper by using given input data sample

        :param model: model object to bind
        :param input_data: input data sample to determine model methods signatures
        :param kwargs: additional information to be used for analysis
        :return: self
        """
        if input_data is None:
            raise ValueError("Input data sample should be specified as 'input_data' key in order to analyze model")

        self.model = model
        self.methods, self.requirements = self._prepare_methods_and_requirements(input_data)
        return self

    def _prepare_methods_and_requirements(self, input_data):
        requirements = Requirements()
        requirements += self._model_requirements()

        arg_type = DatasetAnalyzer.analyze(input_data)
        requirements += arg_type.requirements

        methods = {}
        for exposed, wrapped in self._exposed_methods_mapping().items():
            output_data = self._call_method(wrapped, input_data)
            out_type = DatasetAnalyzer.analyze(output_data)

            methods[exposed] = (wrapped, arg_type, out_type)
            requirements += out_type.requirements
        return methods, requirements

    def unbind(self):
        """
        Unbind model object from this wrapper

        :return: self
        """
        self.model = None
        self.methods = None
        self.requirements = None
        return self

    @property
    def exposed_methods(self) -> typing.Set[str]:
        if self.methods is None:
            raise ValueError('Wrapper {} has no model yet'.format(self))
        return set(self.methods.keys())

    def method_signature(self, name) -> typing.Tuple[DatasetType, DatasetType]:
        """
        Determines input / output types of model wrapper method with given name

        :param name: name of the method to determine input / output types
        :return: input / output type of method with given name
        """
        self._check_method(name)
        _, *signature = self.methods[name]
        return signature

    def call_method(self, name, input_data):
        """
        Calls model wrapper method with given name on given input data

        :param name: name of the method to call
        :param input_data: argument for the method
        :return: call result
        """
        self._check_method(name)
        wrapped, *_ = self.methods[name]
        output_data = self._call_method(wrapped, input_data)
        return output_data

    def _check_method(self, name):
        if self.model is None:
            raise ValueError('Wrapper {} has no model yet'.format(self))
        if name not in self.methods:
            raise ValueError(f"Wrapper '{self}' obj doesn't expose method '{name}'")

    def _call_method(self, wrapped, input_data):
        with switch_curdir(self.curdir):
            if hasattr(self, wrapped):
                return getattr(self, wrapped)(input_data)
            return getattr(self.model, wrapped)(input_data)

    def _model_requirements(self) -> Requirements:
        """
        Should return runtime requirements of bound model.
        By default auto-detects them via Python interpreter internals.
        This is not 100% robust so we recommend to re-implement this method in subclasses.

        :return: :class:`.Requirements` object representing runtime requirements of bound module object
        """
        return get_object_requirements(self)

    @abstractmethod
    def _exposed_methods_mapping(self) -> typing.Dict[str, str]:
        """
        Should return methods exposed by this model wrapper

        :return: methods dict: exposed method name to wrapped method name
        If model wrapper itself has such method then it is going to be called:
        this allows to wrap existing API with your own pre/postprocessing.
        Otherwise, wrapped model object method is going to be called.
        """

    @staticmethod
    def with_model(f):
        """
        Decorator to mark method that only works when model is binded, for example save or predict

        :param f: method
        :return: decorated method
        """

        @wraps(f)
        def wrapper(self, *args, **kwargs):
            if self.model is None:
                raise ValueError('Wrapper {} has no model yet'.format(self))
            return f(self, *args, **kwargs)

        return wrapper

    def __deepcopy__(self, memo):
        cls = type(self)
        obj = object.__new__(cls)
        obj.io = self.io
        obj.model = self.model
        obj.methods = self.methods
        obj.requirements = self.requirements
        obj.curdir = self.curdir
        for field in get_class_fields(cls):
            setattr(obj, field.name, getattr(self, field.name))
        return obj

    def resolve_method(self, method_name=None):
        """Checks if method with this name exists

        :param method_name: name of the method.
        If not provided, this wrapper must have only one method and it will be used"""
        if method_name is None:
            methods = self.exposed_methods
            if len(methods) > 1:
                raise ValueError(f'Please provide one of {methods} as method name')
            method_name = next(iter(methods))
        else:
            self._check_method(method_name)
        return method_name

    def match_methods_by_type(self, input: DatasetType, output: DatasetType) -> typing.List[str]:
        methods = []
        for method_name, (_, method_input, method_output) in self.methods.items():
            if method_input == input and method_output == output:
                methods.append(method_name)
        return methods


class LibModelWrapperMixin(ModelWrapper):
    """
    :class:`.ModelWrapper` mixin which provides model object requirements list consisting of
    PIP packages represented by module objects in `libraries` field.
    """
    libraries = None

    def _model_requirements(self) -> Requirements:
        return Requirements([InstallableRequirement.from_module(lib) for lib in self.libraries])


class WrapperArtifactCollection(ArtifactCollection, Unserializable):
    """
    This is a proxy ArtifactCollection for not persisted artifacts.
    Internally uses :meth:`~algolink.core.objects.ModelWrapper.dump` to create model artifacts

    :param wrapper: :class:`ModelWrapper` instance
    """
    type = '_wrapper'

    def __init__(self, wrapper: ModelWrapper):
        self.wrapper = wrapper

    def materialize(self, path):
        """
        Calls :meth:`~algolink.core.objects.ModelWrapper.dump` to materialize model in path

        :param path: path to materialize model
        """
        with self.wrapper.dump() as art:
            art.materialize(path)

    def bytes_dict(self) -> typing.Dict[str, bytes]:
        """
        Calls :meth:`~algolink.core.objects.wrapper.ModelWrapper.dump` to get model artifacts bytes dict
        :return: dict artifact name -> bytes
        """
        with self.wrapper.dump() as art:
            return art.bytes_dict()

    @contextlib.contextmanager
    def blob_dict(self) -> typing.Iterable[typing.Dict[str, Blob]]:
        """
        Calls :meth:`~algolink.core.objects.wrapper.ModelWrapper.dump` to get model artifacts blob dict

        :return: dict artifact name -> :class:`~algolink.core.objects.artifacts.Blob`
        """
        with self.wrapper.dump() as art, art.blob_dict() as blobs:
            yield blobs


# noinspection PyAbstractClass
class PickleModelIO(ModelIO):
    """
    ModelIO for pickle-able models

    When model is dumped, recursively checks objects if they can be dumped with ModelIO instead of pickling

    So, if you use function that internally calls tensorflow model, this tensorflow model will be dumped with
    tensorflow code and not pickled
    """
    model_filename = 'model.pkl'
    io_ext = '.io'

    @contextlib.contextmanager
    def dump(self, model) -> ArtifactCollection:
        """
        Dumps model artifacts as :class:`~algolink.core.objects.ArtifactCollection`

        :return: context manager with :class:`~algolink.core.objects.ArtifactCollection`
        """
        model_blob, refs = self._serialize_model(model)
        blobs = {self.model_filename: InMemoryBlob(model_blob)}
        artifact_cms = []
        uuids = []

        for uuid, (io, obj) in refs.items():
            blobs[uuid + self.io_ext] = InMemoryBlob(self._serialize_io(io))
            artifact_cms.append(io.dump(obj))
            uuids.append(uuid)

        from .artifacts import _enter_all_cm, _ExitAllCm, _RelativePathWrapper
        additional_artifacts = _enter_all_cm(artifact_cms)
        with _ExitAllCm(artifact_cms):
            additional_artifacts = [_RelativePathWrapper(art, uuid) for art, uuid in zip(additional_artifacts, uuids)]
            yield CompositeArtifactCollection([Blobs(blobs)] + additional_artifacts)

    def load(self, path):
        """
        Loads artifacts into model field

        :param path: path to load from
        """
        refs = {}
        for entry in os.listdir(path):
            if not entry.endswith(self.io_ext):
                continue

            with open(os.path.join(path, entry), 'rb') as f:
                io = self._deserialize_io(f)

            uuid = entry[:-len(self.io_ext)]
            refs[uuid] = io.load(os.path.join(path, uuid))

        with open(os.path.join(path, self.model_filename), 'rb') as f:
            return self._deserialize_model(f, refs)

    @staticmethod
    def _serialize_model(model):
        """
        Helper method to pickle model and get payload and refs

        :return: tuple of payload and refs
        """
        f = BytesIO()
        pklr = _ModelPickler(model, f, recurse=True)
        pklr.dump(model)
        return f.getvalue(), pklr.refs

    @staticmethod
    def _deserialize_model(in_file, refs):
        """
        Helper method to unpickle model from payload and refs

        :param in_file: payload
        :param refs: refs
        :return: unpickled model
        """
        return _ModelUnpickler(refs, in_file).load()

    @staticmethod
    def _serialize_io(io):
        """
        Helper method to serialize object's IO as ref

        :param io: :class:`ModelIO` instance
        :return: ref payload
        """
        io_type = type(io)
        return f'{io_type.__module__}.{io_type.__name__}'.encode('utf-8')

    @staticmethod
    def _deserialize_io(in_file):
        """
        Helper method to deserialize object's IO from ref payload

        :param in_file: ref payload
        :return: :class:`ModelIO` instance
        """
        io_type_full_name = in_file.read().decode('utf-8')
        *mod_name, type_name = io_type_full_name.split('.')
        mod_name, pkg_name = '.'.join(mod_name), '.'.join(mod_name[:-1])
        return import_module(mod_name, pkg_name).__dict__[type_name]()


class _ModelPickler(AlgoLinkPickler):
    """
    A class to pickle model with respect to wrappers of inner objects

    :param model: model object to serialize
    :param args: dill.Pickler args
    :param kwargs: dill.Pickler kwargs
    """

    def __init__(self, model, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.model = model
        self.refs = {}

        # we couldn't import hook and analyzer at top as it leads to circular import failure
        from ..analyzer.model import CallableMethodModelHook, ModelAnalyzer
        known_types = set()
        for hook in ModelAnalyzer.hooks:
            if not isinstance(hook, CallableMethodModelHook):
                known_types.update(hook.valid_types)
        self.known_types = tuple(known_types)

    # pickle "hook" for overriding serialization of objects
    def save(self, obj, save_persistent_id=True):
        """
        Checks if obj has IO.
        If it does, serializes object with :meth:`~algolink.core.objects.wrapper.ModelIO.dump`
        and creates a ref to it. Otherwise, saves object as default pickle would do

        :param obj: obj to save
        :param save_persistent_id:
        :return:
        """
        if obj is self.model:
            # at starting point, follow usual path not to fall into infinite loop
            return super().save(obj, save_persistent_id)

        io = self._get_non_pickle_io(obj)
        if io is None:
            # no non-Pickle IO found, follow usual path
            return super().save(obj, save_persistent_id)

        # found model with non-pickle serialization:
        # replace with `_ExternalRef` stub and memorize IO to serialize model aside later
        obj_uuid = str(uuid4())
        self.refs[obj_uuid] = (io, obj)
        return super().save(_ExternalRef(obj_uuid), save_persistent_id)

    def _get_non_pickle_io(self, obj):
        """
        Checks if obj has non-Pickle IO and returns it

        :param obj: object to check
        :return: non-Pickle :class:`ModelIO` instance or None
        """

        # avoid calling heavy analyzer machinery for "unknown" objects:
        # they are either non-models or callables
        if not isinstance(obj, self.known_types):
            return None

        # we couldn't import analyzer at top as it leads to circular import failure
        from ..analyzer.model import ModelAnalyzer
        try:
            io = ModelAnalyzer._find_hook(obj)._wrapper_factory().io
            return None if isinstance(io, PickleModelIO) else io
        except ValueError:
            # non-model object
            return None


# We couldn't use `AlgoLinkUnpickler` here as it (in fact `dill`) subclasses `Unpickler`
# `Unpickler`, unlike `_Unpickler`, doesn't support `load_build` overriding
class _ModelUnpickler(_Unpickler):
    """
    A class to unpickle model saved with :class:`_ModelPickler`

    :param refs: dict of object uuid -> it's :attr:`ModelWrapper.model`
    :param args: pickle._Unpickler args
    :param kwargs: pickle._Unpickle kwargs
    """
    dispatch = _Unpickler.dispatch.copy()

    def __init__(self, refs, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.refs = refs

    # pickle "hook" for overriding deserialization of objects
    def load_build(self):
        """
        Checks if last builded object is :class:`_ExternalRef` and if it is, swaps it with referenced object
        :return:
        """
        super().load_build()

        # this is the last deserialized object for now
        obj = self.stack[-1]
        if not isinstance(obj, _ExternalRef):
            return

        # replace `_ExternalRef` with a real model it references
        self.stack.pop()
        self.stack.append(self.refs[obj.ref])

    dispatch[pickle.BUILD[0]] = load_build


class _ExternalRef:
    """
    A class to mark objects dumped their own :class:`ModelIO`
    """

    def __init__(self, ref: str):
        self.ref = ref


class CallableMethodModelWrapper(ModelWrapper):
    """
    :class:`ModelWrapper` implementation for functions
    """
    type = 'callable_method'

    def __init__(self):
        super().__init__(PickleModelIO())

    def _exposed_methods_mapping(self) -> typing.Dict[str, str]:
        return {
            'predict': '__call__'
        }



import itertools
import uuid
from io import BytesIO

from algolink.config import Config, Core, Param
from algolink.runtime.interface import Interface
from algolink.runtime.interface.base import InterfaceDescriptor
from algolink.runtime.openapi.spec import create_spec
from algolink.runtime.server import BaseHTTPServer, HTTPServerConfig, MalformedHTTPRequestException
from algolink.utils.fs import current_module_path
from algolink.utils.log import rlogger

from gunicorn.app.base import BaseApplication

current_app = None


def create_executor_function(interface: Interface, method: str):
    """
    Creates a view function for specific interface method

    :param interface: :class:`.Interface` instance
    :param method: method name
    :return: callable view function
    """
    from flask import g, jsonify, request, send_file

    def ef():
        try:
            if request.content_type == 'application/json':
                request_data = BaseHTTPServer._deserialize_json(interface, method, request.json)
            else:
                request_data = dict(itertools.chain(request.form.items(), request.files.items()))

            result = BaseHTTPServer._execute_method(interface, method, request_data, g.algolink_id)

            if isinstance(result, bytes):
                return send_file(BytesIO(result), mimetype='image/png')
            return jsonify(result)
        except MalformedHTTPRequestException as e:
            return jsonify(e.response_body()), e.code()

    ef.__name__ = method

    return ef


def _register_method(app, interface, method_name, signature):
    from flasgger import swag_from

    swag = swag_from(create_spec(method_name, signature, str(Interface), interface.exposed_method_docs(method_name)))
    executor_function = swag(create_executor_function(interface, method_name))
    app.add_url_rule('/' + method_name, method_name, executor_function, methods=['POST'])


def create_interface_routes(app, interface: Interface):
    for method in interface.exposed_methods():
        sig = interface.exposed_method_signature(method)
        rlogger.debug('registering %s with input type %s and output type %s', method, sig.args, sig.output)
        _register_method(app, interface, method, sig)


def create_schema_route(app, interface: Interface):
    from flask import jsonify

    schema = InterfaceDescriptor.from_interface(interface).to_dict()
    rlogger.debug('Creating /interface.json route with schema: %s', schema)
    app.add_url_rule('/interface.json', 'schema', lambda: jsonify(schema))


class FlaskConfig(Config):
    run_flask = Param('run_flask', default='true', parser=bool)


if Core.DEBUG:
    FlaskConfig.log_params()

PREBUILD_PATH = current_module_path('prebuild')
BASE_IMAGE_TEMPLATE = 'algolink/flask:{}'

_host = HTTPServerConfig.host
_port = HTTPServerConfig.port
options = {
    'bind': '{0}:{1}'.format(_host, _port),
    'workers': 2,#multiprocessing.cpu_count(),
}

def prebuild_hook(python_version):
    from ..docker.prebuild import prebuild_image
    prebuild_image(PREBUILD_PATH, BASE_IMAGE_TEMPLATE, python_version)

class APPServer(BaseApplication):
    def __init__(self, app, options=None):
        self.options = options or {}
        self.application = app
        super().__init__()

    def load_config(self):
        config = {key: value for key, value in self.options.items()
                  if key in self.cfg.settings and value is not None}
        for key, value in config.items():
            self.cfg.set(key.lower(), value)
        self.cfg.set("worker_class","gevent")

    def load(self):
        return self.application

class FlaskServer(BaseHTTPServer):
    """
    Flask- and Flasgger-based :class:`.BaseHTTPServer` implementation
    """

    additional_sources = [
        current_module_path('build', 'app.py')  # replace stub in base image
    ]

    additional_options = {'docker': {
        'templates_dir': current_module_path('build'),
        'base_image': lambda python_version: BASE_IMAGE_TEMPLATE.format(python_version),
        'run_cmd': False,  # base image has already specified command
        'prebuild_hook': prebuild_hook
    }}
    def __init__(self,port=9000,serve_method="uwsgi"):
        self.input_port = port
        self.serve_method = serve_method

    def _create_app(self):
        from flasgger import Swagger
        from flask import Flask, g, redirect, request

        app = Flask(__name__)
        app.config['SWAGGER'] = {
            'uiversion': 3,
            'openapi': '3.0.2'
        }
        Swagger(app)

        @app.route('/health')
        def health():
            return 'OK'

        @app.route('/')
        def root():
            return redirect('/apidocs')

        @app.before_request
        def log_request_info():
            g.algolink_id = str(uuid.uuid4())
            app.logger.debug('Headers: %s', request.headers)
            app.logger.debug('Body: %s', request.get_data())

        return app

    def _prepare_app(self, app, interface):
        create_interface_routes(app, interface)
        create_schema_route(app, interface)

    def run(self, interface: Interface):
        """
        Starts flask service

        :param interface: runtime interface to expose via HTTP
        """
        global current_app
        app = self._create_app()
        self._prepare_app(app, interface)

        current_app = app
        if FlaskConfig.run_flask:
            rlogger.debug('Running flask on %s:%s', HTTPServerConfig.host, HTTPServerConfig.port)
            #app.run(HTTPServerConfig.host, HTTPServerConfig.port)
            if self.input_port!=9000:
                _new_port = self.input_port
            else:
                _new_port = HTTPServerConfig.port
            ops = {
                'bind': '{0}:{1}'.format(HTTPServerConfig.host,_new_port),
                'workers': 2,#multiprocessing.cpu_count(),
            }
            if self.serve_method=="uwsgi":
                app.run(HTTPServerConfig.host, HTTPServerConfig.port)
            else:
                APPServer(app, ops).run()
        else:
            rlogger.debug('Skipping direct flask application run')


def main():
    from ..docker.prebuild import prebuild_missing_images
    prebuild_missing_images(PREBUILD_PATH, BASE_IMAGE_TEMPLATE)


if __name__ == '__main__':
    main()



import contextlib
import os
import shutil
import tempfile
from typing import Dict

import tensorflow as tf
from pyjackson.decorators import make_string

from algolink.core.analyzer import TypeHookMixin
from algolink.core.analyzer.model import BindingModelHook
from algolink.core.objects import ModelIO, ModelWrapper
from algolink.core.objects.artifacts import Blobs, LocalFileBlob
from algolink.core.objects.wrapper import FilesContextManager


class TFKerasModelIO(ModelIO):
    """
    :class:`.ModelIO` implementation for Tensorflow Keras models (:class:`tensorflow.keras.Model` objects)
    """

    model_dir_name = 'model.tf'
    ext = '.zip'

    @contextlib.contextmanager
    def dump(self, model) -> FilesContextManager:
        with tempfile.TemporaryDirectory(prefix='algolink_tf_v2') as tmpdir:
            dir_path = os.path.join(tmpdir, self.model_dir_name)

            model.save(dir_path)
            shutil.make_archive(dir_path, 'zip', dir_path)

            yield Blobs({self.model_dir_name + self.ext: LocalFileBlob(dir_path + self.ext)})

    def load(self, path):
        file_path = os.path.join(path, self.model_dir_name + self.ext)

        with tempfile.TemporaryDirectory(prefix='algolink_tf_v2') as tmpdir:
            shutil.unpack_archive(file_path, tmpdir)
            return tf.keras.models.load_model(tmpdir)


class TFKerasModelWrapper(ModelWrapper):
    """
    :class:`.ModelWrapper` implementation for Tensorflow Keras models (:class:`tensorflow.keras.Model` objects)
    """
    def __init__(self):
        super().__init__(TFKerasModelIO())

    def _exposed_methods_mapping(self) -> Dict[str, str]:
        return {
            'predict': 'predict'
        }


@make_string(include_name=True)
class TFKerasModelHook(BindingModelHook, TypeHookMixin):
    """
    Hook for Tensorflow Keras models
    """
    valid_types = [tf.keras.Model]

    def _wrapper_factory(self) -> ModelWrapper:
        """
        Creates :class:`.TFKerasModelWrapper` for Tensorflow Keras model object

        :return: :class:`.TFKerasModelWrapper` instance
        """
        return TFKerasModelWrapper()



import lightgbm as lgb
from pyjackson.core import ArgList
from pyjackson.errors import DeserializationError, SerializationError

from algolink.core.analyzer import TypeHookMixin
from algolink.core.analyzer.dataset import DatasetAnalyzer, DatasetHook
from algolink.core.objects.dataset_type import DatasetType
from algolink.core.objects.requirements import InstallableRequirement, Requirements


class LightGBMDatasetType(DatasetType):
    """
    :class:`.DatasetType` implementation for `lightgbm.Dataset` type

    :param inner: :class:`.DatasetType` instance for underlying data
    """

    real_type = lgb.Dataset

    def __init__(self, inner: DatasetType):
        self.inner = inner

    def is_list(self):
        return self.inner.is_list()

    def list_size(self):
        return self.inner.list_size()

    def get_spec(self) -> ArgList:
        return self.inner.get_spec()

    def serialize(self, instance: lgb.Dataset) -> dict:
        self._check_type(instance, lgb.Dataset, SerializationError)
        return self.inner.serialize(instance.data)

    def deserialize(self, obj: dict) -> lgb.Dataset:
        v = self.inner.deserialize(obj)
        try:
            return lgb.Dataset(v, free_raw_data=False)
        except ValueError:
            raise DeserializationError(f'object: {obj} could not be converted to lightgbm dataset')

    @property
    def requirements(self) -> Requirements:
        return Requirements([InstallableRequirement.from_module(lgb)]) + self.inner.requirements

    @classmethod
    def from_dataset(cls, dataset: lgb.Dataset):
        return cls(DatasetAnalyzer.analyze(dataset.data))


class LightGBMDatasetHook(DatasetHook, TypeHookMixin):
    """
    :class:`.DatasetHook` implementation for `lightgbm.Dataset` type
    """
    valid_types = [lgb.Dataset]

    def process(self, obj, **kwargs) -> DatasetType:
        return LightGBMDatasetType.from_dataset(obj)



import logging
import os
import tempfile
from typing import Any, Callable, Dict, List, Union

import docker.models.images
from docker import errors
from jinja2 import Environment, FileSystemLoader
from pyjackson.utils import get_class_fields

from algolink.build.builder.base import PythonBuildContext, algolink_from_pip
from algolink.build.provider import PythonProvider
from algolink.core.objects.requirements import UnixPackageRequirement
from algolink.utils.log import logger
from algolink.utils.module import get_python_version

from .base import DockerEnv, DockerImage

TEMPLATE_FILE = 'dockerfile.j2'

ALGOLINK_INSTALL_COMMAND = 'pip install algolink=={version};pip install gunicorn;pip install gevent==21.8.0'


def _print_docker_logs(logs, level=logging.DEBUG):
    for log in logs:
        if 'stream' in log:
            logger.log(level, str(log['stream']).strip())
        else:
            logger.log(level, str(log).strip())


class DockerBuildArgs:
    """
    Container for DockerBuild arguments

    :param base_image:  base image for the built image in form of a string or function from python version,
      default: python:{python_version}
    :param python_version: Python version to use, default: version of running interpreter
    :param templates_dir: directory or list of directories for Dockerfile templates, default: ./docker_templates
       - `pre_install.j2` - Dockerfile commands to run before pip
       - `post_install.j2` - Dockerfile commands to run after pip
       - `post_copy.j2` - Dockerfile commands to run after pip and AlgoLink distribution copy
    :param run_cmd: command to run in container, default: sh run.sh
    :param package_install_cmd: command to install packages. Default is apt-get, change it for other package manager
    :param prebuild_hook: callable to call before build, accepts python version. Used for pre-building server images
    """

    def __init__(self, base_image: Union[str, Callable[[str], str]] = None,
                 python_version: str = None, templates_dir: Union[str, List[str]] = None,
                 run_cmd: Union[bool, str] = None, package_install_cmd: str = None,
                 prebuild_hook: Callable[[str], Any] = None):
        self._prebuild_hook = prebuild_hook
        self._package_install_cmd = package_install_cmd
        self._run_cmd = run_cmd
        self._base_image = base_image
        self._python_version = python_version
        self._templates_dir = [] if templates_dir is None else (
            templates_dir if isinstance(templates_dir, list) else [templates_dir])

    @property
    def prebuild_hook(self):
        return self._prebuild_hook

    @property
    def templates_dir(self):
        return self._templates_dir or [os.path.join(os.getcwd(), 'docker_templates')]

    @property
    def package_install_cmd(self):
        return self._package_install_cmd or 'apt-get install -y'

    @property
    def run_cmd(self):
        return self._run_cmd if self._run_cmd is not None else 'sh run.sh'

    @property
    def python_version(self):
        return self._python_version or get_python_version()

    @property
    def base_image(self):
        if self._base_image is None:
            return f'python:{self.python_version}-slim'
        return self._base_image if isinstance(self._base_image, str) else self._base_image(self.python_version)

    def update(self, other: 'DockerBuildArgs'):
        for field in get_class_fields(DockerBuildArgs):
            if field.name == 'templates_dir':
                self._templates_dir += other._templates_dir
            else:
                value = getattr(other, f'_{field.name}')
                if value is not None:
                    setattr(self, f'_{field.name}', value)


class DockerBuildContext(PythonBuildContext):
    """
    PythonBuilder implementation for building docker containers

    :param provider: PythonProvider instance
    :param params: params for docker image to be built
    :param force_overwrite: if false, raise error if image already exists
    :param kwargs: for possible keys, look at :class:`.DockerBuildArgs`
    """

    def __init__(self, provider: PythonProvider, params: DockerImage, force_overwrite=False, **kwargs):
        super().__init__(provider)
        self.params = params
        self.force_overwrite = force_overwrite

        self.args = DockerBuildArgs(python_version=self.provider.get_python_version())

        options = self.provider.get_options()
        if 'docker' in options:
            self.args.update(DockerBuildArgs(**options['docker']))
        self.args.update(DockerBuildArgs(**kwargs))
        self.dockerfile_gen = _DockerfileGenerator(self.args)

    def build(self, env: DockerEnv) -> docker.models.images.Image:
        with tempfile.TemporaryDirectory(prefix='algolink_build_') as tempdir:
            if self.args.prebuild_hook is not None:
                self.args.prebuild_hook(self.args.python_version)
            self._write_distribution(tempdir)
            return self._build_image(tempdir, env)

    def _write_distribution(self, target_dir):
        super()._write_distribution(target_dir)

        logger.debug('Putting Dockerfile to distribution...')
        env = self.provider.get_env()
        logger.debug('Determined environment for running model: %s.', env)
        with open(os.path.join(target_dir, 'Dockerfile'), 'w') as df:
            unix_packages = self.provider.get_requirements().of_type(UnixPackageRequirement)
            dockerfile = self.dockerfile_gen.generate(env, unix_packages)
            df.write(dockerfile)

    def _build_image(self, context_dir, env: DockerEnv) -> docker.models.images.Image:
        tag = self.params.uri
        logger.debug('Building docker image %s from %s...', tag, context_dir)
        with env.daemon.client() as client:
            self.params.registry.login(client)

            if not self.force_overwrite:
                if self.params.exists(client):
                    raise ValueError(f'Image {tag} already exists at {self.params.registry}. '
                                     f'Change name or set force_overwrite=True.')

            else:
                self.params.delete(client)  # to avoid spawning dangling images
            try:
                image, logs = client.images.build(path=context_dir, tag=tag, rm=True)
                logger.info('Built docker image %s', tag)
                _print_docker_logs(logs)

                self.params.registry.push(client, tag)

                return image
            except errors.BuildError as e:
                _print_docker_logs(e.build_log, logging.ERROR)
                raise


class _DockerfileGenerator:
    """
    Class to generate Dockerfile

    :param args: DockerBuildArgs instance
    """

    def __init__(self, args: DockerBuildArgs):
        self.python_version = args.python_version
        self.base_image = args.base_image
        self.templates_dir = args.templates_dir
        self.run_cmd = args.run_cmd
        self.package_install_cmd = args.package_install_cmd

    def generate(self, env: Dict[str, str], packages: List[UnixPackageRequirement] = None):
        """Generate Dockerfile using provided base image, python version and run_cmd

        :param env: dict with environmental variables
        :param packages: list of unix packages to install
        """
        logger.debug('Generating Dockerfile via templates from "%s"...', self.templates_dir)
        j2 = Environment(loader=FileSystemLoader([os.path.dirname(__file__)] + self.templates_dir))
        docker_tmpl = j2.get_template(TEMPLATE_FILE)

        logger.debug('Docker image is using Python version: %s.', self.python_version)
        logger.debug('Docker image is based on "%s".', self.base_image)

        docker_args = {
            'python_version': self.python_version,
            'base_image': self.base_image,
            'run_cmd': self.run_cmd,
            'algolink_install': '',
            'env': env,
            'package_install_cmd': self.package_install_cmd,
            'packages': [p.package_name for p in packages or []]
        }
        algolink_pip = algolink_from_pip()
        if algolink_pip is True:
            import algolink
            docker_args['algolink_install'] = 'RUN ' + ALGOLINK_INSTALL_COMMAND.format(version=algolink.__version__)
        elif isinstance(algolink_pip, str):
            docker_args['algolink_install'] = f"COPY {os.path.basename(algolink_pip)} . " \
                                             f"\n RUN pip install {os.path.basename(algolink_pip)}"

        return docker_tmpl.render(**docker_args)



from typing import Dict

from algolink.runtime.interface import Interface


def merge(ifaces: Dict[str, Interface]) -> Interface:
    """
    Helper to produce composite interface from a number of interfaces.
    Exposes all methods of all given interfaces via given prefixes.

    :param ifaces: dict with (prefix, interface) mappings
    :return: composite interface
    """
    return _MergedInterface(ifaces)


class _MergedInterface(Interface):
    def __init__(self, ifaces):
        exposed = {**self.exposed}
        executors = {**self.executors}
        for pre, iface in ifaces.items():
            for meth in iface.exposed_methods():
                pre_meth = '{}_{}'.format(pre, meth)
                exposed[pre_meth] = iface.exposed_method_signature(meth)
                executors[pre_meth] = self._exec_factory(iface, meth)
        self.exposed = exposed
        self.executors = executors

    @staticmethod
    def _exec_factory(iface, method):
        def _exec(**kwargs):
            return iface.execute(method, kwargs)
        return _exec



from ..core.objects import Model
from ..runtime.command_line import start_runtime
from ..runtime.interface import Interface
from ..runtime.interface.ml_model import ModelLoader, model_interface
from ..runtime.server import Server
from ..utils.importing import module_importable


def run_model_server(model: Model, server: Server = None):
    """
    :func:`.start_runtime` wrapper helper which starts AlgoLink runtime for given model and (optional) server

    :param model: model to start AlgoLink runtime for
    :param server: server to use for AlgoLink runtime, default is a flask-based server
    :return: nothing
    """

    if server is None:
        if module_importable('aiohttp') and module_importable('aiohttp_swagger'):
            from ..ext.aiohttp import AIOHTTPServer
            server = AIOHTTPServer()
        elif module_importable('flask') and module_importable('flasgger'):
            from ..ext.flask import FlaskServer
            server = FlaskServer()
        else:
            raise RuntimeError('You need to install flask and flasgger to use test flask server')

    class DummyLoader(ModelLoader):
        def load(self) -> Interface:
            model.ensure_loaded()
            return model_interface(model)

    start_runtime(DummyLoader(), server)



from abc import ABC

from algolink.core.analyzer import Hook, analyzer_class
from algolink.core.objects.core import Buildable


class BuildableHook(Hook, ABC):
    pass


BuildableAnalyzer = analyzer_class(BuildableHook, Buildable)



import os
from typing import List

from pyjackson import read
from pyjackson.core import Field, Signature

from algolink.core.objects import Model
from algolink.runtime.interface import Interface
from algolink.runtime.interface.base import InterfaceLoader
from algolink.runtime.interface.utils import merge
from algolink.utils.log import rlogger

MODEL_BIN_PATH = 'model_dump'
MODEL_META_PATH = 'model.json'
MODELS_META_PATH = 'models.json'


def model_interface(model_meta: Model):
    """
    Creates an interface from given model with methods exposed by wrapper
    Methods signature is determined via metadata associated with given model.

    :param model_meta: model to create interface for
    :return: instance of :class:`.Interface` implementation
    """

    rlogger.debug('Creating interface for model %s', model_meta)

    class MLModelInterface(Interface):
        def __init__(self, model):
            self.model = model

            exposed = {**self.exposed}
            executors = {**self.executors}

            for name in self.model.exposed_methods:
                in_type, out_type = self.model.method_signature(name)
                exposed[name] = Signature([Field("vector", in_type, False)], Field(None, out_type, False))
                executors[name] = self._exec_factory(name, out_type)

            self.exposed = exposed
            self.executors = executors

        def _exec_factory(self, name, out_type):
            model = self.model

            def _exec(**kwargs):
                input_data = kwargs['vector']
                rlogger.debug('calling %s given %s', name, input_data)
                output_data = model.call_method(name, input_data)
                rlogger.debug('%s returned: %s', name, output_data)
                return out_type.serialize(output_data)

            if model_meta.description is not None:
                _exec.__doc__ = model_meta.description
            return _exec

    return MLModelInterface(model_meta.wrapper)


class ModelLoader(InterfaceLoader):
    """
    Implementation of :class:`.InterfaceLoader` which loads a model via PyJackson and wraps it into an interface
    """

    def load(self) -> Interface:
        meta = read(MODEL_META_PATH, Model)
        meta.wrapper.load(MODEL_BIN_PATH)
        return model_interface(meta)


class MultiModelLoader(InterfaceLoader):
    """
    Implementation of :class:`.InterfaceLoader` which loads a collection of models via PyJackson
    and wraps them into a single interface
    """

    def load(self) -> Interface:
        metas = read(MODELS_META_PATH, List[Model])
        for i, meta in enumerate(metas):
            meta.wrapper.load(os.path.join(MODEL_BIN_PATH, str(i)))
        ifaces = {
            meta.name: model_interface(meta) for meta in metas
        }
        return merge(ifaces)



from typing import Tuple

import tensorflow as tf
from pyjackson.core import ArgList, Field
from pyjackson.errors import DeserializationError, SerializationError

from algolink.core.analyzer import TypeHookMixin
from algolink.core.analyzer.dataset import DatasetHook
from algolink.core.objects.dataset_type import DatasetType, LibDatasetTypeMixin
from algolink.core.objects.typing import ListTypeWithSpec, SizedTypedListType


class TFTensorDatasetType(ListTypeWithSpec, LibDatasetTypeMixin):
    """
    :class:`.DatasetType` implementation for `tensorflow.Tensor` objects
    which converts them to built-in Python lists and vice versa.

    :param shape: shape of `tensorflow.Tensor` objects in dataset
    :param dtype: data type of `tensorflow.Tensor` objects in dataset
    """

    real_type = tf.Tensor
    libraries = [tf]

    def __init__(self, shape: Tuple[int, ...], dtype: str):
        self.shape = (None, ) + shape[1:]
        self.dtype = dtype

    def list_size(self):
        return self.shape[0]

    # noinspection DuplicatedCode
    def _get_subtype(self, shape):
        if len(shape) == 0:
            return self._get_dtype_from_str(self.dtype)
        elif len(shape) == 1:
            subtype = self._get_dtype_from_str(self.dtype)
        else:
            subtype = self._get_subtype(shape[1:])
        return SizedTypedListType(shape[0], subtype)

    @staticmethod
    def _get_dtype_from_str(dtype_str: str):
        known_types = {'float', 'int', 'bool', 'complex', 'str'}
        for known_type in known_types:
            if dtype_str.startswith(known_type):
                return __builtins__[known_type]
        raise ValueError(f'unsupported tensor dtype {dtype_str}')

    def get_spec(self) -> ArgList:
        return [Field(None, self._get_subtype(self.shape[1:]), False)]

    def deserialize(self, obj):
        try:
            ret = tf.convert_to_tensor(obj, dtype=getattr(tf, self.dtype))
        except (ValueError, TypeError):
            raise DeserializationError(f'given object: {obj} could not be converted to tensor '
                                       f'of type: {getattr(tf, self.dtype)}')
        self._check_shape(ret, DeserializationError)
        return ret

    def serialize(self, instance: tf.Tensor):
        self._check_type(instance, tf.Tensor, SerializationError)
        if instance.dtype is not getattr(tf, self.dtype):
            raise SerializationError(f'given tensor is of dtype: {instance.dtype}, '
                                     f'expected: {getattr(tf, self.dtype)}')
        self._check_shape(instance, SerializationError)
        return instance.numpy().tolist()

    def _check_shape(self, tensor, exc_type):
        if tuple(tensor.shape)[1:] != self.shape[1:]:
            raise exc_type(f'given tensor is of shape: {(None,) + tuple(tensor.shape)[1:]}, expected: {self.shape}')


class TFTensorHook(TypeHookMixin, DatasetHook):
    """
    :class:`.DatasetHook` implementation for `tensorflow.Tensor` objects which uses :class:`.TFTensorDatasetType`
    """
    valid_types = [tf.Tensor]

    def process(self, obj, **kwargs) -> DatasetType:
        return TFTensorDatasetType(tuple(obj.shape), obj.dtype.name)



import dill

dill._dill._reverse_typemap['ClassType'] = type


class AlgoLinkPickler(dill.Pickler):
    """Base class for `pickle` serializers in AlgoLink. Based on `dill` library."""
    pass


class AlgoLinkUnpickler(dill.Unpickler):
    """Base class for `pickle` deserializers in AlgoLink. Based on `dill` library."""
    pass



from abc import abstractmethod
from collections import namedtuple
from warnings import warn

from pyjackson import deserialize, serialize

import algolink
from algolink.runtime.interface.base import InterfaceDescriptor, InterfaceMethodDescriptor
from algolink.utils.log import logger


class BaseClient:
    """
    Base class for clients of AlgoLink runtime.

    User method calls are transparently proxied to :class:`~algolink.runtime.interface.base.Interface` deployed on
    :class:`~algolink.runtime.server.base.Server`.
    PyJackson is always used for serialization of inputs and deserialization of outputs.
    """

    def __init__(self):
        self.methods = {}

        interface: InterfaceDescriptor = self._interface_factory()

        if algolink.__version__ != interface.version:
            warn(f"Server AlgoLink version {interface.version}, client AlgoLink version {algolink.__version__}")

        for method in interface.methods:
            self.methods[method.name] = _bootstrap_method(method)

    @abstractmethod
    def _interface_factory(self) -> InterfaceDescriptor:
        """
        Takes interface deployed on server to validate method calls at client side and
        correctly (de)serialize inputs/outputs via PyJackson

        :return: :class:`.InterfaceDescriptor` describing supported methods
        """

        pass  # pragma: no cover

    @abstractmethod
    def _call_method(self, name, args):
        """
        Performs method call at server side

        :param name: name of method to call
        :param args: `dict` of (name, value) mappings for arguments. Values are PyJackson-serialized objects.
        :return: method return value which should be PyJackson deserializable.
        """

        pass  # pragma: no cover

    def __getattr__(self, name):
        if name not in self.methods:
            raise KeyError(f'{name} method is not exposed by server')
        return _MethodCall(self.base_url, self.methods[name], self._call_method)


_Argument = namedtuple('Argument', ('name', 'type'))
_Method = namedtuple('Method', ('name', 'args', 'out_type'))


class _MethodCall:
    def __init__(self, base_url, method: _Method, call_method):
        self.base_url = base_url
        self.method = method
        self.call_method = call_method

    def __call__(self, *args, **kwargs):
        if args and kwargs:
            raise ValueError('Parameters should be passed either in positional or in keyword fashion, not both')
        if len(args) > len(self.method.args) or len(kwargs) > len(self.method.args):
            raise ValueError(f'Too much parameters given, expected: {len(self.method.args)}')

        data = {}
        for i, arg in enumerate(self.method.args):
            obj = None
            if len(args) > i:
                obj = args[i]
            if arg.name in kwargs:
                obj = kwargs[arg.name]
            if obj is None:
                raise ValueError(f'Parameter with name "{arg.name}" (position {i}) should be passed')

            data[arg.name] = serialize(obj, arg.type)

        logger.debug('Calling server method "%s", args: %s ...', self.method.name, data)
        out = self.call_method(self.method.name, data)
        logger.debug('Server call returned %s', out)
        return deserialize(out, self.method.out_type)


def _bootstrap_method(method: InterfaceMethodDescriptor):
    logger.debug('Bootstraping server method "%s" with %s argument(s)...', method.name, len(method.args))
    args = []
    for arg_name, arg_type in method.args.items():
        args.append(_Argument(arg_name, arg_type))

    return _Method(method.name, args, method.out_type)



from typing import List

import xgboost
from pyjackson.core import ArgList, Field
from pyjackson.errors import DeserializationError, SerializationError

from algolink.core.analyzer import TypeHookMixin
from algolink.core.analyzer.dataset import DatasetHook
from algolink.core.objects.dataset_type import DatasetType, LibDatasetTypeMixin
from algolink.ext.numpy.dataset import python_type_from_np_string_repr


class DMatrixDatasetType(LibDatasetTypeMixin):
    """
    :class:`~.DatasetType` implementation for xgboost.DMatrix type

    :param is_from_list: whether DMatrix can be constructed from list
    :param feature_type_names: string representation of feature types
    :param feature_names: list of feature names
    """

    real_type = xgboost.DMatrix
    libraries = [xgboost]

    def __init__(self, is_from_list: bool, feature_type_names: List[str], feature_names: List[str] = None):
        self.is_from_list = is_from_list
        self.feature_type_names = feature_type_names or ['float32' for _ in range(len(feature_names))]
        self.feature_names = feature_names

    def is_list(self):
        return self.is_from_list

    def list_size(self):
        return len(self.feature_names)

    @property
    def feature_types(self):
        return [python_type_from_np_string_repr(t) for t in self.feature_type_names]

    def get_spec(self) -> ArgList:
        return [Field(n, t, False) for n, t in zip(self.feature_names, self.feature_types)]

    def serialize(self, instance: xgboost.DMatrix) -> list:
        """
        Raises an error because there is no way to extract original data from DMatrix
        """
        raise SerializationError('xgboost matrix does not support serialization')

    def deserialize(self, obj: list) -> xgboost.DMatrix:
        try:
            return xgboost.DMatrix(obj)
        except (ValueError, TypeError):
            raise DeserializationError(f'given object: {obj} could not be converted to xgboost matrix')

    @classmethod
    def from_dmatrix(cls, dmatrix: xgboost.DMatrix):
        """
        Factory method to extract :class:`~.DatasetType` from actual xgboost.DMatrix

        :param dmatrix: obj to create :class:`~.DatasetType` from
        :return: :class:`DMatrixDatasetType`
        """
        is_from_list = (dmatrix.feature_names == [f'f{i}' for i in range(dmatrix.num_col())])
        return DMatrixDatasetType(is_from_list, dmatrix.feature_types, dmatrix.feature_names)


class DMatrixHook(DatasetHook, TypeHookMixin):
    """
    :class:`.DatasetHook` implementation for `xgboost.DMatrix` objects
    """
    valid_types = [xgboost.DMatrix]

    def process(self, obj: xgboost.DMatrix, **kwargs) -> DatasetType:
        return DMatrixDatasetType.from_dmatrix(obj)



from typing import Dict, List, Type

from pyjackson.utils import (Field, Signature, get_class_fields, get_collection_internal_type, get_mapping_types,
                             is_collection, is_generic, is_mapping)

from algolink.core.objects.dataset_type import BytesDatasetType, PrimitiveDatasetType
from algolink.core.objects.typing import TypeWithSpec

BUILTIN_TYPES: Dict[Type, str] = {
    int: 'integer', float: 'number', str: 'string', None: 'null', bool: 'boolean',  # list:'array', dict:'object'
}


def make_object(properties: List[Field] = None,
                arbitrary_properties_type: Type = None,
                has_default=False, default=None):  # TODO nullable
    """
    Converts object type described as list of fields to OpenAPI schema definition

    :param properties: fields of object
    :param arbitrary_properties_type: (optional) required type for properties which are not specified in `properties`
    :param has_default: specifies whether given type has default value
    :param default: specifies default value for given type
    :return: dict with OpenAPI schema definition
    """

    result = {
        'type': 'object',
    }
    if properties is not None:
        result['properties'] = {a.name: _field_to_schema(a) for a in properties}
        required = [a.name for a in properties if not a.has_default]
        if required:
            result['required'] = required
    if arbitrary_properties_type is not None:
        result['additionalProperties'] = type_to_schema(arbitrary_properties_type)
    if has_default:
        result['default'] = default

    return result


def make_array(item_type: Type, minimum_size=None, maximum_size=None, has_default=False, default=None):
    """
    Converts array type described as type of its items and range of possible sizes to OpenAPI schema definition

    :param item_type: type of items in array
    :param minimum_size: minimal possible size of array
    :param maximum_size: maximal possible size of array
    :param has_default: specifies whether given type has default value
    :param default: specifies default value for given type
    :return: dict with OpenAPI schema definition
    """

    result = {
        'type': 'array',
        'items': type_to_schema(item_type)
    }
    if minimum_size is not None:
        result['minItems'] = minimum_size
    if maximum_size is not None:
        result['maxItems'] = maximum_size
    if has_default:
        result['default'] = default
    return result


def _field_to_schema(field: Field):
    return type_to_schema(field.type, field.has_default, field.default)


def type_to_schema(field_type, has_default=False, default=None):
    """
    Facade method converting arbitrary type to OpenAPI schema definitions.
    Has special support for builtins, collections and instances of :class:`.TypeWithSpec` subclasses.

    :param field_type: type to generate schema for
    :param has_default: specifies whether given type has default value
    :param default: specifies default value for given type
    :return: dict with OpenAPI schema definition
    """
    if field_type in BUILTIN_TYPES:
        result = {'type': BUILTIN_TYPES[field_type]}
        if has_default:
            result['default'] = default
        return result

    elif is_generic(field_type):
        if is_collection(field_type):
            return make_array(get_collection_internal_type(field_type), has_default=has_default, default=default)
        elif is_mapping(field_type):
            kt, vt = get_mapping_types(field_type)
            if kt != str:
                raise ValueError('Only string keys supported')
            return make_object(arbitrary_properties_type=vt, has_default=has_default, default=default)

    elif issubclass(field_type, TypeWithSpec):
        if issubclass(field_type, PrimitiveDatasetType):
            return type_to_schema(field_type.to_type, has_default, default)

        # noinspection PyArgumentList
        spec = field_type.get_spec()
        # noinspection PyArgumentList
        is_list = field_type.is_list()
        # noinspection PyArgumentList
        list_size = field_type.list_size()
        if is_list:
            item_types = set(a.type for a in spec)
            if len(item_types) > 1:
                raise ValueError('All items must be of same type for {}'.format(field_type))
            item_type = next(iter(item_types))
            return make_array(item_type, list_size, list_size)
        else:
            return make_object(spec)
        # return schema_from_typespec(field)
    else:
        return make_object(get_class_fields(field_type))


def create_spec(method_name: str, signature: Signature, name: str, docs: str):
    """
    Generates OpenAPI schema definition for given method

    :param method_name: name of method
    :param signature: types of arguments and type of return value
    :param name: name of the interface
    :param docs: docs for method
    :return: dict with OpenAPi schema definition
    """

    error_def = {
        'type': 'object',
        'properties': {
            'ok': {'type': 'boolean'},
            'error': {'type': 'string'}
        }
    }
    value_args = [a for a in signature.args if not issubclass(a.type, BytesDatasetType)]
    file_args = [a for a in signature.args if issubclass(a.type, BytesDatasetType)]

    if issubclass(signature.output.type, BytesDatasetType):
        good_response = {"description": "successful response",
                         'content': {"multipart/form-data": {
                             'type': 'string',
                             'format': 'binary'
                         }}}
    else:
        response_def = {
            'type': 'object',
            'properties': {
                'ok': {'type': 'boolean'},
                'data': type_to_schema(signature.output.type, False, None)
            }
        }
        good_response = {"description": "successful response",
                         "content": {"application/json": {"schema": response_def}}}

    if value_args and file_args:
        raise ValueError('Both JSON and file data at the same time is not supported')

    if value_args:
        json_body_def = {
            'type': 'object',
            'properties': {
                a.name: _field_to_schema(a) for a in value_args
            }
        }
        request_body = {'required': True,
                        "content": {"application/json": {'schema': json_body_def}}}
    elif file_args:
        files_body_def = {
            'type': 'object',
            'properties': {
                a.name: {'type': 'string', 'format': 'binary'} for a in file_args
            }
        }
        request_body = {'required': True,
                        "content": {"multipart/form-data": {'schema': files_body_def}}}
    else:
        request_body = {'required': False}

    bad_response = {"description": "incorrect request",
                    "content": {"application/json": {"schema": error_def}}}

    summary = f"Calls '{method_name}' method on {name}."
    if docs is not None:
        summary += f' Method description: {docs}'
    return {
        "summary": summary,
        "requestBody": request_body,
        "responses": {
            "200": good_response,
            "400": bad_response
        }
    }



from abc import abstractmethod
from typing import Dict

from algolink.core.objects.artifacts import ArtifactCollection
from algolink.core.objects.requirements import Requirements
from algolink.runtime.interface import InterfaceLoader
from algolink.runtime.server import Server
from algolink.utils.module import get_python_version


class ProviderBase:
    """Base class for providers"""

    @abstractmethod
    def get_sources(self) -> Dict[str, str]:
        """Abstract method for text files"""
        pass  # pragma: no cover

    @abstractmethod
    def get_artifacts(self) -> ArtifactCollection:
        """Abstract method for binaries"""
        pass  # pragma: no cover

    @abstractmethod
    def get_env(self) -> Dict[str, str]:
        """Abstract method for environment variables"""
        pass  # pragma: no cover

    @abstractmethod
    def get_options(self) -> Dict[str, str]:
        """Abstract method for additional build options"""
        pass  # pragma: no cover


SERVER_ENV = 'ALGOLINK_SERVER'
LOADER_ENV = 'ALGOLINK_LOADER'


class PythonProvider(ProviderBase):
    """Provider for python-based builds. Includes python version and requirements

    :param server: Server instance to build with
    :param loader: InterfaceLoader instance to build with
    :param debug: Whether to run image in debug mode
    """

    def __init__(self, server: Server, loader: InterfaceLoader, debug: bool = False):
        self.debug = debug
        self.server = server
        self.loader = loader

    def get_python_version(self):
        """Returns current python version"""
        return get_python_version()

    @abstractmethod
    def get_requirements(self) -> Requirements:
        """Abstract method for python requirements"""
        pass  # pragma: no cover

    def get_env(self) -> Dict[str, str]:
        """Get env variables for image"""
        envs = {
            LOADER_ENV: self.loader.classpath,
            SERVER_ENV: self.server.classpath,
            'ALGOLINK_RUNTIME': 'true'
        }
        if self.debug:
            envs['ALGOLINK_DEBUG'] = 'true'

        envs.update(self.server.additional_envs)

        modules = set(self.get_requirements().modules)

        from algolink.ext import ExtensionLoader
        extensions = ExtensionLoader.loaded_extensions.keys()
        used_extensions = [e.module for e in extensions if all(r in modules for r in e.reqs)]
        if len(used_extensions) > 0:
            envs['ALGOLINK_EXTENSIONS'] = ','.join(used_extensions)
        return envs

    def get_options(self) -> Dict[str, str]:
        return self.server.additional_options



from abc import abstractmethod
from collections import Iterable
from typing import Any, Optional

from pyjackson.core import Unserializable
from pyjackson.decorators import type_field

from algolink.core.analyzer.dataset import DatasetAnalyzer
from .base import AlgoLinkParams
from .dataset_type import DatasetType


class AbstractDataset(Unserializable):
    """ABC for Dataset objects

    :param dataset_type: DatasetType instance for the data in the Dataset"""

    def __init__(self, dataset_type: DatasetType):
        self.dataset_type = dataset_type
        self.writer = None
        self.reader = None

    @abstractmethod
    def iterate(self) -> Iterable:
        """Abstract method to iterate through data"""

    @abstractmethod
    def get(self):
        """Abstract method to get data object"""

    @abstractmethod
    def get_writer(self):
        """Returns writer for this dataset. Defaults to dataset_type.get_writer()"""
        return self.writer or self.dataset_type.get_writer()

    @abstractmethod
    def get_reader(self):
        """Returns reader for this dataset. Defaults to dataset_type.get_reader()"""
        return self.reader or self.dataset_type.get_reader()


class Dataset(AbstractDataset):
    """Wrapper for dataset objects

    :param data: raw dataset
    :param dataset_type: DatasetType of the raw data"""

    def __init__(self, data: Any, dataset_type: DatasetType):
        super().__init__(dataset_type)
        self.data = data

    def iterate(self) -> Iterable:
        return iter(self.data)

    def get(self):
        return self.data

    @classmethod
    def from_object(cls, data):
        """Creates Dataset instance from raw data object"""
        return cls(data, DatasetAnalyzer.analyze(data))

    def to_inmemory_source(self) -> 'InMemoryDatasetSource':
        """Returns :class:`.InMemoryDatasetSource` with this dataset"""
        return InMemoryDatasetSource(self)


@type_field('type')
class DatasetSource(AlgoLinkParams):
    """Class that represents a source that can produce a Dataset

    :param dataset_type: DatasetType of contained dataset"""
    is_dynamic = False

    def __init__(self, dataset_type: DatasetType):
        self.dataset_type = dataset_type

    @abstractmethod
    def read(self) -> Dataset:
        """Abstract method that must return produced Dataset instance"""
        raise NotImplementedError()

    def cache(self):
        """Returns :class:`.CachedDatasetSource` that will cache data on the first read"""
        return CachedDatasetSource(self)


class CachedDatasetSource(DatasetSource):
    """Wrapper that will cache the result of underlying source on the first read

    :param source: underlying DatasetSource"""

    def __init__(self, source: DatasetSource):
        super().__init__(source.dataset_type)
        self.source = source
        self._cache: Optional[Dataset] = None

    def read(self) -> Dataset:
        if self._cache is None:
            self._cache = self.source.read()
        return self._cache

    def cache(self):
        return self


class InMemoryDatasetSource(CachedDatasetSource, Unserializable):
    """DatasetSource that holds existing dataset inmemory

    :param dataset: Dataset instance to hold"""

    def __init__(self, dataset: Dataset):
        super().__init__(DatasetSource(dataset.dataset_type))
        self._cache = dataset



from flasgger import Swagger, swag_from
from flask import Flask, jsonify

app = Flask(__name__)
Swagger(app)


@app.route('/', methods=['GET'])
@swag_from({'responses': {'200': {'content': {'application/json': {'schema': {
    'type': 'object', 'properties': {'hello': {'type': 'string'}}}}}}}})
def hello():
    return jsonify({'hello': 'world'}), 200


__all__ = ['app']



from abc import abstractmethod
from typing import Generator, Type

from algolink.core.objects import Image, RuntimeEnvironment, RuntimeInstance


class RunnerBase:
    @abstractmethod
    def instance_type(self) -> Type[RuntimeInstance.Params]:
        """
        :return: subtype of :class:`.RuntimeInstance.Params` supported by this runner
        """

    @abstractmethod
    def create_instance(self, name: str, **kwargs) -> RuntimeInstance.Params:
        """
        Creates new runtime instance on given name and args

        :param name: name of instance to use
        :return: created :class:`.RuntimeInstance.Params` subclass instance
        """

    @abstractmethod
    def run(self, instance: RuntimeInstance.Params, image: Image.Params, env: RuntimeEnvironment.Params, **kwargs):
        """
        Runs given image on given environment with params given by instance

        :param instance: instance params to use for running
        :param image: image to base instance on
        :param env: environment to run on
        """

    @abstractmethod
    def is_running(self, instance: RuntimeInstance.Params, env: RuntimeEnvironment.Params, **kwargs) -> bool:
        """
        Checks that given instance is running on given environment

        :param instance: instance to check running of
        :param env: environment to check running on
        :return: "is running" flag
        """

    @abstractmethod
    def stop(self, instance: RuntimeInstance.Params, env: RuntimeEnvironment.Params, **kwargs):
        """
        Stops running of given instance on given environment

        :param instance: instance to stop running of
        :param env: environment to stop running on
        """

    @abstractmethod
    def logs(self, instance: RuntimeInstance.Params, env: RuntimeEnvironment.Params, **kwargs) \
            -> Generator[str, None, None]:
        # TODO :param stream: stream or just print latest
        """
        Exposes logs produced by given instance while running on given environment

        :param instance: instance to expose logs for
        :param env: environment to expose logs from

        :return: generator of log strings or string with logs
        """

    @abstractmethod
    def instance_exists(self, instance: RuntimeInstance.Params, env: RuntimeEnvironment.Params, **kwargs) -> bool:
        """Checks if instance exists in environment

        :param instance: instance params to check
        :param env: environment to check in
        :return: boolean flag
        """

    @abstractmethod
    def remove_instance(self, instance: RuntimeInstance.Params, env: RuntimeEnvironment.Params, **kwargs):
        """
        Removes instance

        :param instance: instance params to remove
        :param env: environment to remove from
        """



import os
import warnings
from typing import List, Optional, Union

from pyjackson import dumps
from pyjackson.decorators import cached_property

from algolink.build.provider.utils import BuildableWithServer
from algolink.core.analyzer import CanIsAMustHookMixin
from algolink.core.analyzer.buildable import BuildableHook
from algolink.build.provider.ml_model import MLModelProvider, read
from algolink.core.objects import ArtifactCollection, Model, Requirements, Task
from algolink.core.objects.artifacts import _RelativePathWrapper, CompositeArtifactCollection
from algolink.core.objects.core import WithMetadataRepository
from algolink.runtime.interface.ml_model import MODEL_BIN_PATH, MODELS_META_PATH
from algolink.runtime.server import Server
from algolink.utils.module import get_object_requirements
from algolink.utils.log import logger


class MLModelMultiProvider(MLModelProvider):
    """Provider to put multiple models in one service

    :param models: List of Model instances
    :param server: Server instance to build with
    :param debug: Debug for instance"""

    def __init__(self, models: List[Model], server: Server, debug: bool = False):
        from algolink.runtime.interface.ml_model import MultiModelLoader
        super(MLModelProvider, self).__init__(server, MultiModelLoader(), debug)
        self.models: List[Model] = models

    @cached_property
    def _requirements(self) -> Requirements:
        """Union of server, loader and all models requirements"""
        return (get_object_requirements(self.server) +
                get_object_requirements(self.loader) +
                sum((model.requirements for model in self.models), Requirements()))

    def get_requirements(self):
        """Returns union of server, loader and all models requirements"""
        return self._requirements

    def get_sources(self):
        """Returns models meta file and custom requirements"""
        return {
            MODELS_META_PATH: dumps([model.without_artifacts() for model in self.models]),
            **self._get_sources(),
            **{os.path.basename(f): read(f) for f in self.server.additional_sources}
        }

    def get_artifacts(self) -> ArtifactCollection:
        """Returns binaries of models artifacts"""
        # TODO additional server binaries
        return CompositeArtifactCollection([
            _RelativePathWrapper(m.artifact_any, os.path.join(MODEL_BIN_PATH, str(i)))
            for i, m in enumerate(self.models)
        ])

    def get_python_version(self):
        versions = [model.params.get(Model.PYTHON_VERSION) for model in self.models]
        if len(set(versions)) > 1:
            logger.warn('Models in MultModelProvider have varying python versions in requirements')
        return max(versions)


class MultiModelBuildable(BuildableWithServer, WithMetadataRepository):
    def __init__(self, model_ids: List[Union[int, Model]], server_type: str, debug: bool = False):
        if len(model_ids) == 0:
            raise ValueError('model_ids must contain at least one model')
        super().__init__(server_type)
        if isinstance(model_ids[0], int):
            self.model_ids = model_ids
            self.models_cache = None
        else:
            self.models_cache = model_ids
            self.model_ids = [m.id for m in model_ids]
            self.bind_meta_repo(self.models_cache[0]._meta)
        self.debug = debug

    @property
    def task(self) -> Optional[Task]:
        tasks = [m.task for m in self.models]
        if len(set(t.id for t in tasks)) != 1:
            warnings.warn(f'Ambiguious task for buildable {self}')
        return tasks[0]

    @property
    def models(self) -> List[Model]:
        if self.models_cache is None:
            self._check_meta(False)
            self.models_cache = [self._meta.get_model_by_id(mid) for mid in self.model_ids]
        return self.models_cache

    def get_provider(self) -> MLModelMultiProvider:
        return MLModelMultiProvider(self.models, self.server, self.debug)


class BuildableMultiModelHook(BuildableHook, CanIsAMustHookMixin):

    def must_process(self, obj) -> bool:
        return isinstance(obj, list) and all(isinstance(o, Model) for o in obj)

    def process(self, obj, **kwargs):
        server = kwargs.get('server')  # TODO ???
        debug = kwargs.get('debug', False)
        return MultiModelBuildable(obj, server.type, debug)



from algolink.build.builder.base import BuilderBase
from algolink.build.helpers import validate_kwargs
from algolink.core.objects.core import Buildable
from algolink.ext.docker.build_context import DockerBuildArgs, DockerBuildContext

from .base import DockerEnv, DockerImage


class DockerBuilder(BuilderBase):
    """Builder implementation to build docker images"""
    @validate_kwargs
    def create_image(self, name: str, environment: DockerEnv, tag: str = 'latest', repository: str = None,
                     **kwargs) -> DockerImage:
        return DockerImage(name, tag, repository, environment.registry)

    @validate_kwargs(allowed_funcs=[DockerBuildArgs.__init__])
    def build_image(self, buildable: Buildable, image: DockerImage, environment: DockerEnv,
                    force_overwrite=False, **kwargs):
        context = DockerBuildContext(buildable.get_provider(), image, force_overwrite=force_overwrite, **kwargs)
        docker_image = context.build(environment)
        image.image_id = docker_image.id

    @validate_kwargs
    def delete_image(self, image: DockerImage, environment: DockerEnv, force=False, **kwargs):
        with environment.daemon.client() as client:
            image.delete(client, force, **kwargs)

    @validate_kwargs
    def image_exists(self, image: DockerImage, environment: DockerEnv, **kwargs) -> bool:
        with environment.daemon.client() as client:
            return image.exists(client)
